from __future__ import annotations
import jax.numpy as np
from jax.scipy.signal import convolve
from jax import vmap
from equinox import tree_at
from zodiax import Base
from abc import ABC, abstractmethod
from dLux.utils.units import convert_angular, convert_cartesian
import dLux


__all__ = ["PointSource", "MultiPointSource", "ArrayDistribution",
           "BinarySource", "PointExtendedSource", "PointAndExtendedSource"]


Array = np.ndarray


########################
### Abstract Classes ###
########################
class Source(Base, ABC):
    """
    Base class for source objects. The idea of these source classes is to allow
    an arbitrary parametrisation of the underlying astrophyical objects. Each
    source object requires a normalise(), format_inputs(), and model() methods,
    along with the regular getter and setter methods.

    The normalise() method should return a new instance of the class with all
    the appropriate attribues normalised (ie spectral weights, resolved source
    distribution).

    The format_inputs() method should return the relevant wavelengths, weights,
    positions etc, that are correctly formatted to be used by the model()
    method. This should primarily interface with the specrum object, and should
    call the normalise() method.

    The model() method should return a single psf of the source. This is done
    inside of the source object so that each class can be arbitrarily
    parameterised as required. This method should call the format_inputs()
    method.

    A series of parameterisations have been proided to model a series of
    different astrophysical objects.

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : Array, photons
        The flux of the object.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    name : str
        The name for this object.
    """
    position : Array
    flux     : Array
    spectrum : Spectrum
    name     : str


    def __init__(self     : Source,
                 position : Array    = None,
                 flux     : Array    = None,
                 spectrum : Spectrum = None,
                 name     : str      = 'Source') -> Source:
        """
        Constructor for the Source class.

        Parameters
        ----------
        position : Array, radians = None
            The (x, y) on-sky position of this object.
        flux : Array, photons = None
            The flux of the object.
        spectrum : Spectrum = None
            The spectrum of this object, represented by a Spectrum object.
        name : str = 'Source'
            The name for this object.
        """
        self.position = np.asarray(position, dtype=float)
        self.flux     = np.asarray(flux,     dtype=float)
        self.spectrum = spectrum
        self.name     = name

        # Input position checking
        if position is not None:
            assert self.position.ndim == 1, \
            ("position must be a 1d array.")
            assert self.position.shape == (2,), \
            ("positions must be shape (2,), ie (x, y).")

        # Input flux checking
        if flux is not None:
            assert self.flux.shape == (), \
            ("flux must be a scalar, (shape == ()).")

        # Input spectrum checking
        assert isinstance(self.spectrum, dLux.spectrums.Spectrum), \
        ("Spectrum must be dLux Spectrum object.")

        # Input name checking
        assert isinstance(self.name, str), "Name must be a string."
    

    def __getattr__(self : Source, key : str) -> Any:
        """
        Getter method for the spectrum object.

        Parameters
        ----------
        key : str
            The key to get from the spectrum object.

        Returns
        -------
        value : Any
            The value of the key.
        """
        if hasattr(self.spectrum, key):
            return getattr(self.spectrum, key)
        else:
            raise AttributeError(f"{self.name} has no attribute {key}.")


    def get_flux(self : Source) -> Array:
        """
        Getter method for the flux.

        Returns
        -------
        flux : Array, photons
            The flux of the object.
        """
        return self.flux


    def get_position(self : Source) -> Array:
        """
        Getter method for the position.

        Returns
        -------
        position : Array, radians
            The (x, y) on-sky position of this object.
        """
        return self.position


    def get_wavelengths(self : Source) -> Array:
        """
        Getter method for the source internal wavelengths.

        Returns
        -------
        wavelengths : Array, meters
            The array of wavelengths at which the spectrum is defined.
        """
        return self.spectrum.get_wavelengths()


    def get_weights(self : Source) -> Array:
        """
        Getter method for the source internal weights.

        Returns
        -------
        weights : Array
            The relative weights of each wavelength.
        """
        return self.spectrum.get_weights()


    def normalise(self : Source) -> Source:
        """
        Method for returning a new normalised source object.

        Returns
        -------
        source : Source
            The normalised source object.
        """
        normalised_spectrum = self.spectrum.normalise()
        return tree_at(
            lambda source : source.spectrum, self, normalised_spectrum)
    

    def summary(self            : Source, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name} has no summary method yet.")
    

    def display(self            : Source, 
                figsize         : tuple = (6, 3),
                dpi             : int = 120,
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> None:
        """
        Displays a plot of the wavefront amplitude and opd or phase.

        Parameters
        ----------
        figsize : tuple = (6, 3)
            The size of the figure to display.
        cmap : str = 'inferno'
            The colour map to use.
        dpi : int = 120
            The resolution of the figure.
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.
        """
        self.spectrum.normalise().display(figsize=figsize, dpi=dpi, 
            angular_units=angular_units, cartesian_units=cartesian_units)


    def model(self      : Source,
              optics    : Optics,
              detector  : Detector = None,
              filter_in : Filter   = None) -> Array:
        """
        Method to model the psf of the point source through the optics.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source objects.
        detector : Detector = None
            The detector object that is observing the psf.
        filter_in : Filter = None
            The filter through which the source is being observed.

        Returns
        -------
        psf : Array
            The psf of the source source modelled through the optics.
        """
        # Normalise and get parameters
        self        = self.normalise()
        wavelengths = self.get_wavelengths()
        weights     = self.get_weights()
        position    = self.get_position()
        flux        = self.get_flux()

        # Get filter throughput
        if filter_in is not None:
            raise NotImplementedError("Filter modelling is under development.")

        # Vmap propagator & model
        propagator = vmap(optics.propagate_mono, in_axes=(0, None))
        psfs = weights[:, None, None] * propagator(wavelengths, position)
        psf = flux * psfs.sum(0)

        # Apply detector if supplied
        return psf if detector is None else detector.apply_detector(psf)


class ResolvedSource(Source, ABC):
    """
    Base class for resolved source objects. This simply extends the base Source
    class by implementing an abstract get_distribution() method and a concrete
    model() method.

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : Array, photons
        The flux of the object.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    name : str
        The name for this object.
    """


    @abstractmethod
    def get_distribution(self): # pragma: no cover
        """
        Abstract method for returning the distribution of the resolved source.

        Returns
        -------
        distribution : Array
            The distribution of the resolved source
        """
        pass


    def model(self      : Source,
              optics    : Optics,
              detector  : Detector = None,
              filter_in : Filter   = None) -> Array:
        """
        Method to model the psf of the source through the optics. Implements a
        basic convolution with the psf and source distribution.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source objects.
        detector : Detector = None
            The detector object that is observing the psf.
        filter_in : Filter = None
            The filter through which the source is being observed.

        Returns
        -------
        psf : Array
            The psf of the source modelled through the optics
        """
        # Normalise and get parameters
        self         = self.normalise()
        wavelengths  = self.get_wavelengths()
        weights      = self.get_weights()
        position     = self.get_position()
        flux         = self.get_flux()
        distribution = self.get_distribution()

        # Get filter throughput
        if filter_in is not None:
            raise NotImplementedError("Filter modelling is under development.")

        # Vmap propagator & model
        propagator = vmap(optics.propagate_mono, in_axes=(0, None))
        psfs = weights[:, None, None] * propagator(wavelengths, position)
        psf = convolve(flux * psfs.sum(0), distribution, mode='same')

        # Apply detector if supplied
        return psf if detector is None else detector.apply_detector(psf)


class RelativeFluxSource(Source, ABC):
    """
    Abstract class that extend the methods of Source to allow for binary-object
    sources to be parameterised by their relative flux. Classes that inherit
    from this class must instantiate a contrast attribute.

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : Array, photons
        The flux of the object.
    contrast : Array
        The contrast ratio between the two sources.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    name : str
        The name for this object.
    """
    contrast : Array


    def __init__(self      : Source,
                 contrast : Array,
                 **kwargs) -> Source:
        """
        Constructor for the RelativeFluxSource class.

        Parameters
        ----------
        contrast : Array
            The contrast ratio between the two sources.
        """
        super().__init__(**kwargs)
        self.contrast = np.asarray(contrast, dtype=float)

        # Input contrast checking
        assert self.contrast.shape == (), \
        ("Flux ratio must be a scalar, (shape == ()).")


    def get_flux(self : Source) -> Array:
        """
        Getter method for the fluxes. This paramterieses the source such that
        flux refers to the mean_flux and contrast is defined as the ratio of
        the flux of the first entry divided by the second entry.

        Returns
        -------
        flux : Array, photons
            The flux (flux1, flux2) of the binary object.
        """
        flux_A = 2 * self.contrast * self.flux / (1 + self.contrast)
        flux_B = 2 * self.flux / (1 + self.contrast)
        return np.array([flux_A, flux_B])


class RelativePositionSource(Source, ABC):
    """
    Abstract class that extend the methods of Source to allow for binary-object
    sources to be parameterised by their relative position. Classes that
    inherit from this class must instantiate a separation attribute.

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : Array, photons
        The flux of the object.
    separation : Array, radians
        The separation of the two sources in radians.
    position_angle : Array, radians
        The field angle between the two sources measure from the positive
        x axis.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    name : str
        The name for this object.
    """
    separation     : Array
    position_angle : Array


    def __init__(self           : Source,
                 separation     : Array,
                 position_angle : Array,
                 **kwargs) -> Source:
        """
        Constructor for the RelativePositionSource class.

        Parameters
        ----------
        separation : Array, radians
            The separation of the two sources in radians.
        position_angle : Array, radians
            The field angle between the two sources measure from the positive
            x axis.
        """
        super().__init__(**kwargs)
        self.separation     = np.asarray(separation,     dtype=float)
        self.position_angle = np.asarray(position_angle, dtype=float)

        assert self.separation.shape == (), "Separation must be a scalar, \
        (shape == ())."

        assert self.position_angle.shape == (), "Field angle must be a scalar, \
        (shape == ())."


    def get_position(self : Source) -> Array:
        """
        Getter method for the position.

        Returns
        -------
        position : Array, radians
            The ((x, y), (x, y)) on-sky position of this object.
        """
        r, phi = self.separation/2, self.position_angle
        sep_vec = np.array([r*np.sin(phi), r*np.cos(phi)])
        return np.array([self.position + sep_vec,
                         self.position - sep_vec])


########################
### Concrete Classes ###
########################
class PointSource(Source):
    """
    Concrete Class for unresolved point source objects.

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : Array, photons
        The flux of the object.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    name : str
        The name for this object.
    """


    def __init__(self        : Source,
                 position    : Array    = np.array([0., 0.]),
                 flux        : Array    = np.array(1.),
                 spectrum    : Spectrum = None,
                 wavelengths : Array    = None,
                 name        : str      = 'PointSource') -> Source:
        """
        Constructor for the PointSource class.

        Parameters
        ----------
        position : Array, radians = np.array([0., 0.])
            The (x, y) on-sky position of this object.
        flux : Array, photons = np.array(1.)
            The flux of the object.
        spectrum : Spectrum = None
            The spectrum of this object, represented by a Spectrum object.
        wavelengths : Array, meters = None
            The array of wavelengths at which the spectrum is defined.
        name : str = 'PointSource'
            The name for this object.
        """
        # Check spectrum & wavelengths
        assert spectrum is not None or wavelengths is not None, \
        ("Either spectrum or wavelengths must be specified.")
        if spectrum is not None:
            assert wavelengths is None, \
            ("wavelengths can not be specified if spectrum is specified.")
        else:
            spectrum = dLux.spectrums.ArraySpectrum(wavelengths)
        super().__init__(position, flux, spectrum, name=name)


    def summary(self            : Source, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        position = np.array_str(convert_angular(self.position, 'radians', 
                                            angular_units), precision=sigfigs)
        spectrum_str = self.spectrum.summary(angular_units, cartesian_units,
                                                sigfigs)
        flux = self.flux
        return (f"{self.name}: A Point Source with flux {flux:.{sigfigs}} "
                f"photons at offset from the optical axis by {position} "
                f"{angular_units} with spectrum\n  {spectrum_str}")


class MultiPointSource(Source):
    """
    Concrete Class for multiple unresolved point source objects.

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky positions of these sources.
    flux : Array, photons
        The fluxes of the sources.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
        Every source in this class will have an identical spectrum.
    name : str
        The name for this object.
    """


    def __init__(self        : Source,
                 position    : Array,
                 flux        : Array    = None,
                 spectrum    : Spectrum = None,
                 wavelengths : Array    = None,
                 name        : str      = 'MultiPointSource') -> Source:
        """
        Constructor for the MultiPointSource class.

        Parameters
        ----------
        position : Array, radians
            The ((x0, y0), (x1, y1), ...) on-sky positions of these sourcese.
        flux : Array, photons = None
            The fluxes of the sources.
        spectrum : Spectrum = None
            The spectrum of this object, represented by a Spectrum object.
            Every source in this class will have an identical spectrum.
        wavelengths : Array, meters = None
            The array of wavelengths at which the spectrum is defined.
        name : str = 'MultiPointSource'
            The name for this object.
        """
        # Check spectrum & wavelengths
        assert spectrum is not None or wavelengths is not None, \
        ("Either spectrum or wavelengths must be specified.")
        if spectrum is not None:
            assert wavelengths is None, \
            ("wavelengths can not be specified if spectrum is specified.")
        else:
            spectrum = dLux.spectrums.ArraySpectrum(wavelengths)

        # Only call super, not __init__ since we are overwriting all the attrs
        super().__init__(spectrum=spectrum, name=name)

        self.position = np.asarray(position, dtype=float)

        # Input position checking
        assert self.position.ndim == 2, \
        ("position must be a 2d array.")
        assert self.position.shape[-1] == 2, \
        ("positions must be shape (nstars, 2), ie [(x0, y0), (x1, y1), ...].")

        # Get flux
        self.flux = np.ones(len(self.positions)) if flux is None else \
                    np.asarray(flux, dtype=float)

        # Input flux checking
        assert self.flux.ndim == 1, \
        ("flux must be a 1d array.")

        # Ensure same dimensionality
        assert self.flux.shape[0] == self.position.shape[0], ("position and "
        "flux must have the same length leading dimension, ie nstars")


    def model(self      : Source,
              optics    : Optics,
              detector  : Detector = None,
              filter_in : Filter   = None) -> Array:
        """
        Method to model the psf of the point source through the optics.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source objects.
        detector : Detector = None
            The detector object that is observing the psf.
        filter_in : Filter = None
            The filter through which the source is being observed.

        Returns
        -------
        psf : Array
            The psf of the source source modelled through the optics.
        """
        # Normalise and get parameters
        self        = self.normalise()
        wavelengths = self.get_wavelengths()
        weights     = self.get_weights()
        positions   = self.get_position()
        fluxes      = self.get_flux()

        # Get filter throughput
        if filter_in is not None:
            raise NotImplementedError("Filter modelling is under development.")

        # Vmap propagator
        source_propagator = vmap(optics.propagate_mono, in_axes=(0, None))
        propagator = vmap(source_propagator, in_axes=(None, 0))

        # Model Psf
        psfs = propagator(wavelengths, positions)
        psfs *= weights[None, :, None, None] * fluxes[:, None, None, None]
        psf = psfs.sum((0, 1))

        # Apply detector if supplied
        return psf if detector is None else detector.apply_detector(psf)
    

    def summary(self            : Source, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        position = np.array_str(convert_angular(self.position, 'radians', 
                                            angular_units), precision=sigfigs)
        spectrum_str = self.spectrum.summary(angular_units, cartesian_units,
                                                sigfigs)
        n = len(self.flux)
        return (f"{self.name}: {n} Point Sources with fluxes "
                f"{self.flux:.{sigfigs}} photons at offsets from the optical "
                f"axis by {position} {angular_units} with spectrum\n  "
                f"{spectrum_str}")


class ArrayDistribution(ResolvedSource):
    """
    A class for modelling resolved sources that parameterise their resolved
    component using an array of intensities.

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : Array, photons
        The flux of the object.
    distribution : Array
        The array of intensities respresenting the resolved source.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    name : str
        The name for this object.
    """
    distribution : Array


    def __init__(self         : Source,
                 position     : Array    = np.array([0., 0.]),
                 flux         : Array    = np.array(1.),
                 distribution : Array    = np.ones((3, 3)),
                 spectrum     : Spectrum = None,
                 wavelengths  : Array    = None,
                 name         : str      = 'ArrayDistribution',
                 **kwargs) -> Source:
        """
        Constructor for the ArrayDistribution class.

        Parameters
        ----------
        position : Array, radians = np.array([0., 0.])
            The (x, y) on-sky position of this object.
        flux : Array, photons = np.array(1.)
            The flux of the object.
        distribution : Array = np.ones((3, 3))
            The array of intensities respresenting the resolved source.
        spectrum : Spectrum = None
            The spectrum of this object, represented by a Spectrum object.
        wavelengths : Array, meters = None
            The array of wavelengths at which the spectrum is defined.
        name : str = 'ArrayDistribution'
            The name for this object.
        """
        # Check spectrum & wavelengths
        assert spectrum is not None or wavelengths is not None, \
        ("Either spectrum or wavelengths must be specified.")
        if spectrum is not None:
            assert wavelengths is None, \
            ("wavelengths can not be specified if spectrum is specified.")
        else:
            spectrum = dLux.spectrums.ArraySpectrum(wavelengths)

        super().__init__(position, flux, spectrum, name=name, **kwargs)
        distribution = np.asarray(distribution, dtype=float)
        distribution = np.maximum(distribution, 0.)
        self.distribution = distribution/distribution.sum()

        # Input checking
        assert self.distribution.ndim == 2, \
        ("distribution must be a 2d array.")
        assert len(self.distribution) > 0, \
        ("Length of distribution must be greater than 1.")


    def get_distribution(self : Source) -> Array:
        """
        Getter method for the source distribution.

        Returns
        -------
        distribution : Array, intensity
            The distribution of the source intensity.
        """
        return self.distribution


    def normalise(self : Source) -> Source:
        """
        Method for returning a new source object with a normalised total
        spectrum and source distribution.

        Returns
        -------
        source : Source
            The source object with the normalised spectrum and distribution.
        """
        normalised_spectrum = self.spectrum.normalise()
        distribution_floor = np.maximum(self.distribution, 0.)
        normalised_distribution = distribution_floor/distribution_floor.sum()
        return tree_at(
            lambda source : (source.spectrum, source.distribution), self, \
                            (normalised_spectrum, normalised_distribution))
    

    def summary(self            : Source, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        position = np.array_str(convert_angular(self.position, 'radians', 
                                            angular_units), precision=sigfigs)
        spectrum_str = self.spectrum.summary(angular_units, cartesian_units,
                                                sigfigs)
        return (f"{self.name}: A Resolved Source with flux "
                f"{self.flux:.{sigfigs}} photons at offset from the optical "
                f"axis by {position} {angular_units} with spectrum\n  "
                f"{spectrum_str}")


class BinarySource(RelativePositionSource, RelativeFluxSource):
    """
    A parameterised binary source.

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : Array, photons
        The mean flux of the sources.
    separation : Array, radians
        The separation of the two sources in radians.
    position_angle : Array, radians
        The position angle between the two sources measured clockwise from
        the vertical axis.
    contrast : Array
        The contrast ratio between the two sources.
    spectrum : Spectrum
        The spectrum of this object, represented by a CombinedSpectrum object.
    name : str
        The name for this object.
    """


    def __init__(self           : Source,
                 position       : Array    = np.array([0., 0.]),
                 flux           : Array    = np.array(1.),
                 separation     : Array    = None,
                 position_angle : Array    = np.pi/2,
                 contrast       : Array    = np.array(1.),
                 spectrum       : Spectrum = None,
                 wavelengths    : Array    = None,
                 name           : str      = 'BinarySource') -> Source:
        """
        Parameters
        ----------
        position : Array, radians = np.array([0., 0.])
            The (x, y) on-sky position of this object.
        flux : Array, photons = np.array(1.)
            The mean flux of the sources.
        separation : Array, radians = None
            The separation of the two sources in radians.
        position_angle : Array, radians = np.pi/2
            The position angle between the two sources measured clockwise from
            the vertical axis.
        contrast : Array = np.array(1.)
            The contrast ratio between the two sources.
        spectrum : CombinedSpectrum = None
            The spectrum of this object, represented by a CombinedSpectrum.
        wavelengths : Array, meters = None
            The array of wavelengths at which the spectrum is defined.
        name : str = 'BinarySource'
            The name for this object.
        """
        # Check spectrum & wavelengths
        assert spectrum is not None or wavelengths is not None, \
        ("Either spectrum or wavelengths must be specified.")
        if spectrum is not None:
            assert wavelengths is None, \
            ("wavelengths can not be specified if spectrum is specified.")
            assert isinstance(spectrum, dLux.CombinedSpectrum), \
            ("The input spectrum must be a CombinedSpectrum object.")
        else:
            spectrum = dLux.spectrums.CombinedSpectrum(wavelengths)

        # Check separation
        assert separation is not None, ("separation must be provided.")

        super().__init__(position=position,
                         flux=flux,
                         spectrum=spectrum,
                         separation=separation,
                         position_angle=position_angle,
                         contrast=contrast,
                         name=name)


    def model(self      : Source,
              optics    : Optics,
              detector  : Detector = None,
              filter_in : Filter   = None) -> Array:
        """
        Method to model the psf of the point source through the optics.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source objects.
        detector : Detector = None
            The detector object that is observing the psf.
        filter_in : Filter = None
            The filter through which the source is being observed.

        Returns
        -------
        psf : Array
            The psf of the source source modelled through the optics.
        """
        # Normalise and get parameters
        self        = self.normalise()
        wavelengths = self.get_wavelengths()[0]
        weights     = self.get_weights()
        positions   = self.get_position()
        fluxes      = self.get_flux()

        # Get filter throughput
        if filter_in is not None:
            raise NotImplementedError("Filter modelling is under development.")

        # Vmap propagator
        source_propagator = vmap(optics.propagate_mono, in_axes=(0, None))
        propagator = vmap(source_propagator, in_axes=(None, 0))

        # Model Psf
        psfs = propagator(wavelengths, positions)
        psfs *= weights[:, :, None, None] * fluxes[:, None, None, None]
        psf = psfs.sum((0, 1))

        # Apply detector if supplied
        return psf if detector is None else detector.apply_detector(psf)
    

    def summary(self            : Source, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        position = np.array_str(convert_angular(self.position, 'radians', 
                                            angular_units), precision=sigfigs)
        separation = convert_angular(self.separation, 'radians', angular_units)
        position_angle = convert_angular(self.position_angle, 'radians', 
                                         angular_units)
        spectrum_str = self.spectrum.summary(angular_units, cartesian_units,
                                                sigfigs)
        return (f"{self.name}: A Binary Source with mean flux "
                f"{self.flux:.{sigfigs}} photons and constrast {self.contrast} "
                f"offset from the optical axis by {position} "
                f"{angular_units} with separation {separation:.{sigfigs}} "
                f"{angular_units}, position angle {position_angle:.{sigfigs}} "
                f"{angular_units} and spectrum\n  {spectrum_str}")


class PointExtendedSource(RelativeFluxSource, ArrayDistribution):
    """
    A class for modelling a point source and a resolved source that is defined
    relative to the point source. An example would be an unresolved star with
    a resolved dust shell or debris disk. These two objects share the same
    spectra but have their fluxes defined by flux (the mean flux) and the flux
    ratio (contrast) between the point source and resolved distribution. The
    resolved component is defined by an array (ie this class inherits from
    ArrayDistribution).

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : Array, photons
        The mean flux of the point and resolved source.
    distribution : Array
        The array of intensities respresenting the resolved source.
    contrast : Array
        The contrast ratio between the point source and the resolved
        source.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    name : str
        The name for this object.
    """


    def __init__(self         : Source,
                 position     : Array    = np.array([0., 0.]),
                 flux         : Array    = np.array(1.),
                 distribution : Array    = np.ones((3, 3)),
                 contrast     : Array    = np.array(1.),
                 spectrum     : Spectrum = None,
                 wavelengths  : Array    = None,
                 name         : str      = 'PointExtendedSource') -> Source:
        """
        Parameters
        ----------
        position : Array, radians = np.array([0., 0.])
            The (x, y) on-sky position of this object.
        flux : Array, photons = np.array(1.)
            The mean flux of the point and resolved source.
        distribution : Array = np.ones((3, 3))
            The array of intensities respresenting the resolved source.
        contrast : Array = np.array(1.)
            The contrast ratio between the point source and the resolved
            source.
        spectrum : Spectrum = None
            The spectrum of this object, represented by a Spectrum object.
        wavelengths : Array, meters = None
            The array of wavelengths at which the spectrum is defined.
        name : str = 'PointExtendedSource'
            The name for this object.
        """
        # Check spectrum & wavelengths
        assert spectrum is not None or wavelengths is not None, \
        ("Either spectrum or wavelengths must be specified.")
        if spectrum is not None:
            assert wavelengths is None, \
            ("wavelengths can not be specified if spectrum is specified.")
        else:
            spectrum = dLux.spectrums.Spectrum(wavelengths)

        super().__init__(position=position, flux=flux, spectrum=spectrum, \
                         distribution=distribution, contrast=contrast, \
                         name=name)


    def model(self      : Source,
              optics    : Optics,
              detector  : Detector = None,
              filter_in : Filter   = None) -> Array:
        """
        Method to model the psf of the source through the optics. Implements a
        basic convolution with the psf and source distribution, while also
        modelling the single point source psf.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the soource objects.
        detector : Detector = None
            The detector object that is observing the psf.
        filter_in : Filter = None
            The filter through which the source is being observed. Default is 
            None which is uniform throughput.

        Returns
        -------
        psf : Array
            The psf of the source modelled through the optics.
        """
        # Normalise and get parameters
        self         = self.normalise()
        wavelengths  = self.get_wavelengths()
        weights      = self.get_weights()
        position     = self.get_position()
        fluxes       = self.get_flux()
        distribution = self.get_distribution()

        # Get filter throughput
        if filter_in is not None:
            raise NotImplementedError("Filter modelling is under development.")

        # Vmap propagator
        propagator = vmap(optics.propagate_mono, in_axes=(0, None))

        # Model psfs
        psfs = propagator(wavelengths, position)
        single_psf = (weights[:, None, None] * psfs).sum(0)
        point_psf = fluxes[0] * single_psf
        extended_psf = fluxes[1] * single_psf
        convolved = convolve(extended_psf, distribution, mode='same')
        psf = convolved + point_psf

        # Apply detector if supplied
        return psf if detector is None else detector.apply_detector(psf)
    

    def summary(self            : Source, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        position = np.array_str(convert_angular(self.position, 'radians', 
                                            angular_units), precision=sigfigs)
        spectrum_str = self.spectrum.summary(angular_units, cartesian_units,
                                                sigfigs)
        return (f"{self.name}: A Point and Resolved Source with mean flux "
                f"{self.flux:.{sigfigs}} photons and constrast "
                f"{self.contrast:.{sigfigs}} offset from the optical axis by "
                f"{position} {angular_units} and spectrum\n  "
                f"{spectrum_str}")


class PointAndExtendedSource(RelativeFluxSource, ArrayDistribution):
    """
    A class for modelling a point source and a resolved source that is defined
    relative to the point source, but with its own spectra. An example would be
    an unresolved quasar within a resolved galaxy. These two objects have
    independent spectra but have their fluxes defined by flux (the mean flux)
    and the flux ratio (contrast) between the point source and resolved
    distribution. The resolved component is defined by an array (ie this class
    inherits from ArrayDistribution).

    Attributes
    ----------
    position : Array, radians
        The (x, y) on-sky position of this object.
    flux : Array, photons
        The mean flux of the point and resolves source.
    distribution : Array
        The array of intensities respresenting the resolved source.
    contrast : Array
        The contrast ratio between the point source and the resolved
        source.
    spectrum : Spectrum
        The spectrum of this object, represented by a Spectrum object.
    name : str
        The name for this object.
    """


    def __init__(self         : Source,
                 position     : Array    = np.array([0., 0.]),
                 flux         : Array    = np.array(1.),
                 distribution : Array    = np.ones((3, 3)),
                 contrast     : Array    = np.array(1.),
                 spectrum     : Spectrum = None,
                 wavelengths  : Array    = None,
                 name         : str      = 'PointAndExtendedSource') -> Source:
        """
        Parameters
        ----------
        position : Array, radians = np.array([0., 0.])
            The (x, y) on-sky position of this object.
        flux : Array, photons = np.array(1.)
            The flux of the object.
        distribution : Array = np.ones((3, 3))
            The array of intensities respresenting the resolved source.
        contrast : Array = np.array(1.)
            The contrast ratio between the point source and the resolved
            source.
        spectrum : CombinedSpectrum = None
            The spectrum of this object, represented by a CombinedSpectrum.
        wavelengths : Array, meters = None
            The array of wavelengths at which the spectrum is defined.
        name : str = 'PointAndExtendedSource'
            The name for this object.
        """
        # Check spectrum & wavelengths
        assert spectrum is not None or wavelengths is not None, \
        ("Either spectrum or wavelengths must be specified.")
        if spectrum is not None:
            assert wavelengths is None, \
            ("wavelengths can not be specified if spectrum is specified.")
            assert isinstance(spectrum, dLux.CombinedSpectrum), \
            ("The input spectrum must be a CombinedSpectrum object.")
        else:
            spectrum = dLux.spectrums.CombinedSpectrum(wavelengths)

        super().__init__(position=position, flux=flux, spectrum=spectrum, \
                         distribution=distribution, contrast=contrast, \
                         name=name)


    def model(self      : Source,
              optics    : Optics,
              detector  : Detector = None,
              filter_in : Filter   = None) -> Array:
        """
        Method to model the psf of the source through the optics. Implements a
        basic convolution with the psf and source distribution, while also
        modelling the single point source psf. Applied a different spectrum to
        the point source and resolved source.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source objects.
        detector : Detector = None
            The detector object that is observing the psf.
        filter_in : Filter = None
            The filter through which the source is being observed. Default is
            None which is uniform throughput.

        Returns
        -------
        psf : Array
            The psf of the source modelled through the optics.
        """
        # Normalise and get parameters
        self         = self.normalise()
        wavelengths  = self.get_wavelengths()[0]
        weights      = self.get_weights()
        position     = self.get_position()
        fluxes       = self.get_flux()
        distribution = self.get_distribution()

        # Get filter throughput
        if filter_in is not None:
            raise NotImplementedError("Filter modelling is under development.")

        # Vmap propagator
        propagator = vmap(optics.propagate_mono, in_axes=(0, None))

        # Model psfs
        psfs = propagator(wavelengths, position)
        point_psf    = fluxes[0] * (weights[0, :, None, None] * psfs).sum(0)
        extended_psf = fluxes[1] * (weights[1, :, None, None] * psfs).sum(0)
        convolved = convolve(extended_psf, distribution, mode='same')
        psf = convolved + point_psf

        # Apply detector if supplied
        return psf if detector is None else detector.apply_detector(psf)
    

    def summary(self            : Source, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        position = np.array_str(convert_angular(self.position, 'radians', 
                                            angular_units), precision=sigfigs)
        spectrum_str = self.spectrum.summary(angular_units, cartesian_units,
                                                sigfigs)
        return (f"{self.name}: A Point and Resolved Source with mean flux "
                f"{self.flux} photons and constrast {self.contrast} offset "
                f"from the optical axis by {position} "
                f"{angular_units} and spectrum\n  {spectrum_str}")



from __future__ import annotations
import jax.numpy as np
import pytest
import dLux
from jax import config
config.update("jax_debug_nans", True)

class TestArraySpectrum():
    """
    Tests the ArraySpectrum class.
    """

    def test_constructor(self, create_array_spectrum : callable) -> None:
        """
        Tests the constructor.
        """
        # Test string inputs
        with pytest.raises(ValueError):
            create_array_spectrum(weights="")

        # Test zero dimension input
        with pytest.raises(AssertionError):
            create_array_spectrum(weights=5.)

        # Test zero length input
        with pytest.raises(AssertionError):
            create_array_spectrum(weights=[])


    def test_get_weights(self, create_array_spectrum : callable) -> None:
        """
        Tests the get_weights method.
        """
        spectrum = create_array_spectrum()
        assert (spectrum.get_weights() == spectrum.weights).all()


    def test_normalise(self, create_array_spectrum : callable) -> None:
        """
        Tests the normalise method.
        """
        new_weights = np.arange(10)
        new_spectrum = create_array_spectrum().set('weights',
                                                    new_weights).normalise()
        assert np.allclose(new_spectrum.weights.sum(), 1.)



class TestPolynomialSpectrum():
    """
    Tests the PolynomialSpectrum class.

    Note this does not test the .normalise() method becuase it does not
    normalise the coefficients, instead the .get_weights() returns a normalised
    weights.
    """


    def test_constructor(self, create_polynomial_spectrum : callable) -> None:
        """
        Tests the constructor.
        """
        # Test string inputs
        with pytest.raises(ValueError):
            create_polynomial_spectrum(coefficients="")

        # Test zero dimension input
        with pytest.raises(AssertionError):
            create_polynomial_spectrum(coefficients=5.)

        # Test zero length input
        with pytest.raises(AssertionError):
            create_polynomial_spectrum(coefficients=[])

        # Test nan inputs
        with pytest.raises(AssertionError):
            create_polynomial_spectrum(coefficients=[np.nan])

        # Test infinite inputs
        with pytest.raises(AssertionError):
            create_polynomial_spectrum(coefficients=[np.inf])

        create_polynomial_spectrum()

    def test_get_weights(self, create_polynomial_spectrum : callable) -> None:
        """
        Tests the normalisation of the get_weights method.
        """
        assert np.allclose(create_polynomial_spectrum().get_weights().sum(), 1.)


class TestCombinedSpectrum():
    """
    Tests the CombinedSpectrum class
    """


    def test_constructor(self, create_combined_spectrum : callable) -> None:
        """
        Test the constructor.
        """
        # Wavelengths Testing
        # Test string inputs
        with pytest.raises(ValueError):
            create_combined_spectrum(wavelengths="")

        # Test zero dimension input
        with pytest.raises(AssertionError):
            create_combined_spectrum(wavelengths=5.)

        # Test zero length input
        with pytest.raises(AssertionError):
            create_combined_spectrum(wavelengths=[])

        # Weights Testing
        # Test string inputs
        with pytest.raises(ValueError):
            create_combined_spectrum(weights="")

        # Test zero dimension input
        with pytest.raises(AssertionError):
            create_combined_spectrum(weights=5.)

        # Test zero length input
        with pytest.raises(AssertionError):
            create_combined_spectrum(weights=[])


    def test_normalise(self, create_combined_spectrum : callable) -> None:
        """
        Tests the normalise method.
        """
        new_weights = np.tile(np.arange(10), (2, 1))
        new_spectrum = create_combined_spectrum().set('weights',
                                                    new_weights).normalise()
        assert np.allclose(new_spectrum.weights.sum(1), 1.)



from __future__ import annotations
import jax.numpy as np
import pytest
import dLux
from jax import config
config.update("jax_debug_nans", True)

Array = np.ndarray


class TestApplyPixelResponse(object):
    """
    Tests the ApplyPixelResponse class.
    """


    def test_constructor(self, create_pixel_response: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_pixel_response(pixel_response=np.ones(1))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_pixel_response(pixel_response=np.array(1.))

        # Test wrong dims
        with pytest.raises(AssertionError):
            create_pixel_response(pixel_response=np.ones((1, 1, 1)))

        # Test functioning
        create_pixel_response()


    def test_call(self, create_pixel_response: callable) -> None:
        """
        Tests the __call__ method.
        """
        image = np.ones((16, 16))
        create_pixel_response()(image)


class TestApplyJitter(object):
    """
    Tests the ApplyJitter class.
    """


    def test_constructor(self, create_jitter: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_jitter(sigma=np.ones(1))

        # Test functioning
        create_jitter()


    def test_call(self, create_jitter: callable) -> None:
        """
        Tests the __call__ method.
        """
        image = np.ones((16, 16))
        create_jitter()(image)


class TestApplySaturation(object):
    """
    Tests the ApplySaturation class.
    """


    def test_constructor(self, create_saturation: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_saturation(saturation=np.ones(1))

        # Test functioning
        create_saturation()


    def test_call(self, create_saturation: callable) -> None:
        """
        Tests the __call__ method.
        """
        image = np.ones((16, 16))
        create_saturation()(image)


class TestAddConstant(object):
    """
    Tests the AddConstant class.
    """


    def test_constructor(self, create_constant: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_constant(value=np.ones(1))

        # Test functioning
        create_constant()


    def test_call(self, create_constant: callable) -> None:
        """
        Tests the __call__ method.
        """
        image = np.ones((16, 16))
        create_constant()(image)


class TestIntegerDownsample(object):
    """
    Tests the IntegerDownsample class.
    """


    def test_constructor(self, create_integer_downsample: callable) -> None:
        """
        Tests the constructor.
        """
        # Test functioning
        create_integer_downsample()


    def test_call(self, create_integer_downsample: callable) -> None:
        """
        Tests the __call__ method.
        """
        image = np.ones((16, 16))
        create_integer_downsample()(image)


class TestRotate(object):
    """
    Tests the Rotate class.
    """


    def test_constructor(self, create_rotate_detector: callable) -> None:
        """
        Tests the constructor.
        """
        # Test wrong dims
        with pytest.raises(AssertionError):
            create_rotate_detector(angle=np.ones(1))

        # Test functioning
        create_rotate_detector()


    def test_call(self, create_rotate_detector: callable) -> None:
        """
        Tests the __call__ method.
        """
        # Test regular rotation
        image = np.ones((16, 16))
        create_rotate_detector()(image)

        # Test fourier
        with pytest.raises(NotImplementedError):
            create_rotate_detector(fourier=True)(image)



from __future__ import annotations
from abc import ABC, abstractmethod
import jax.numpy as np
from equinox import tree_at
from zodiax import Base
from jax import vmap
from dLux.utils.units import convert_cartesian
from dLux.utils.helpers import spectrum_plot


__all__ = ["ArraySpectrum", "PolynomialSpectrum", "CombinedSpectrum"]


Array = np.ndarray


class Spectrum(Base, ABC):
    """
    Abstract base class for arbitrary spectral parametrisations.

    Attributes
    ----------
    wavelengths : Array, meters
        The array of wavelengths at which the spectrum is defined.
    name : str
        The name of the spectrum.
    """
    name        : str
    wavelengths : Array


    def __init__(self        : Spectrum,
                 wavelengths : Array,
                 name        : str = 'Spectrum') -> Spectrum:
        """
        Constructor for the Spectrum class.

        Parameters
        ----------
        wavelengths : Array, meters
            The array of wavelengths at which the spectrum is defined.
        name : str = 'Spectrum'
            The name of the spectrum.
        """
        self.name = name
        self.wavelengths = np.asarray(wavelengths, dtype=float)

        # Input checking
        assert self.wavelengths.ndim == 1, \
        ("Wavelengths must be a 1d array.")
        assert len(self.wavelengths) > 0, \
        ("Length of wavelengths must be greater than 1.")


    def get_wavelengths(self : Spectrum) -> Array:
        """
        Getter method for the wavelengths.

        Returns
        -------
        wavelengths : Array, meters
            The array of wavelengths at which the spectrum is defined.
        """
        return self.wavelengths


    @abstractmethod
    def normalise(self : Spectrum) -> Spectrum: # pragma: no cover
        """
        Abstract method to normalise the spectrum. Must be overwitten by child
        classes.
        """
        return


    @abstractmethod
    def get_weights(self : Spectrum) -> Array: # pragma: no cover
        """
        Abstract getter method for the weights. Must be overwritten by child
        classes.
        """
        return
    

    def summary(self            : Spectrum, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return f"{self.name} layer has no summary method yet."
    

    def display(self            : Spectrum, 
                figsize         : tuple = (6, 3),
                dpi             : int = 120,
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> None:
        """
        Displays a plot of the wavefront amplitude and opd or phase.

        Parameters
        ----------
        figsize : tuple = (6, 3)
            The size of the figure to display.
        cmap : str = 'inferno'
            The colour map to use.
        dpi : int = 120
            The resolution of the figure.
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.
        """
        wavelengths = convert_cartesian(self.wavelengths, "meters", 
                                        cartesian_units)
        spectrum_plot(wavelengths, self.get_weights(), figsize=figsize,
                      cartesian_units=cartesian_units, dpi=dpi)


class ArraySpectrum(Spectrum):
    """
    A Spectrum class that interally parametersises the spectrum via arrays (ie
    wavelengths and weights)

    Attributes
    ----------
    wavelengths : Array, meters
        The array of wavelengths at which the spectrum is defined.
    weights : Array
        The relative weights of each wavelength.
    name : str
        The name of the spectrum.
    """
    weights : Array


    def __init__(self        : Spectrum,
                 wavelengths : Array,
                 weights     : Array = None,
                 name        : str = 'ArraySpectrum') -> Spectrum:
        """
        Constructor for the ArraySpectrum class.

        Parameters
        ----------
        wavelengths : Array, meters
            The array of wavelengths at which the spectrum is defined.
        weights : Array = None
            The relative weights of each wavelength. Defaults to uniform
            spectrum. Weights are automatically normalised to a sum of 1.
        name : str = 'ArraySpectrum'
            The name of the spectrum.
        """
        super().__init__(wavelengths, name)
        weights = np.ones(len(self.wavelengths))/len(wavelengths) \
                                    if weights is None else weights
        weights = np.asarray(weights, dtype=float)
        self.weights = weights/np.sum(weights)

        # Input checks
        assert self.weights.ndim == 1, \
        ("weights must be a 1d array.")
        assert len(self.weights) > 0, \
        ("Length of weights must be greater than 1.")
        assert len(self.wavelengths) == len(self.weights), \
        ("wavelengths and weights must have the same length.")


    def get_weights(self : Spectrum) -> Array:
        """
        Getter method for the weights.

        Returns
        -------
        weights : Array
            The relative weights of each wavelength.
        """
        return self.weights


    def normalise(self : Spectrum) -> Spectrum:
        """
        Method for returning a new spectrum object with a normalised total
        spectrum.

        Returns
        -------
        spectrum : Specturm
            The spectrum object with the normalised spectrum.
        """
        total_power = self.get_weights().sum()
        norm_weights = self.get_weights()/total_power
        return tree_at(lambda spectrum: spectrum.weights, self, norm_weights)
    

    def summary(self            : Spectrum, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Applies a spectrum with weights stored as an "
                "array.")


class PolynomialSpectrum(Spectrum):
    """
    Implements a generic polynomial spectrum. This is likely not needed and
    will probably just be turned into LinearSpectrum in the future.

    This implements a polynomial as follows:
    f(x) = c0 + c1*x + c2*x^2 + ... + cn*x^n

    Attributes
    ----------
    wavelengths : Array, meters
        The array of wavelengths at which the spectrum is defined.
    degree : int
        The degree of the polynomial.
    coefficients : Array
        The array of polynomial coefficient values.
    name : str
        The name of the spectrum.
    """
    degree       : int # Just a helper
    coefficients : Array


    def __init__(self         : Spectrum,
                 wavelengths  : Array,
                 coefficients : Array,
                 name        : str = 'PolynomialSpectrum') -> Spectrum:
        """
        Constructor for the PolynomialSpectrum class.

        Parameters
        ----------
        wavelengths : Array, meters
            The array of wavelengths at which the spectrum is defined.
        coefficients : Array
            The array of polynomial coefficient values.
        name : str = 'PolynomialSpectrum'
            The name of the spectrum.
        """
        super().__init__(wavelengths, name)

        self.coefficients = np.asarray(coefficients, dtype=float)

        assert self.coefficients.ndim == 1, "Coefficients must be a 1d array."
        assert not np.isnan(self.coefficients).any(), \
        ("Coefficients must not be nan.")
        assert not np.isinf(self.coefficients).any(), \
        ("Coefficients must be not be infinite.")

        # Input checks
        assert self.coefficients.ndim == 1, \
        ("coefficients must be a 1d array.")
        assert len(self.coefficients) > 0, \
        ("Length of coefficients must be greater than 1.")

        self.degree = int(len(coefficients) - 1)


    def get_weights(self : Spectrum) -> Array:
        """
        Gets the relative spectral weights by evalutating the polynomial
        function at the internal wavelengths. This automaically normalises
        the weights to have unitary amplitude.

        Returns
        -------
        weights : Array
            The normalised relative weights of each wavelength.
        """
        generate_polynomial = vmap(lambda wavelength :
                                np.array([self.coefficients[i] * wavelength**i 
                                for i in range(len(self.coefficients))]).sum())
        weights = generate_polynomial(self.wavelengths)
        return weights/weights.sum()


    def normalise(self : Spectrum) -> Spectrum:
        """
        This method currently does nothing becuase solving for normalised
        polynomial coefficients is difficut and the get_weights() method
        already returns normalised weights.

        Returns:
        spectrum : Specturm
            The unmodified spectrum object
        """
        return self
    

    def summary(self            : Spectrum, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Applies a spectrum with weights parametried by "
                f"a polynomial of degree {self.degree}.")


class CombinedSpectrum(ArraySpectrum):
    """
    Implements a combined spectrum, in order to have only a single spectrum
    object for parameterised sources such as binary stars.

    TODO: Expand to be arbitrary number of spectrums, store only a single
    wavelengths array and tile it on the get_wavelengths method. (ie make work
    for MultiPointSource)

    Attributes
    ----------
    wavelengths : Array, meters
        The (nspectra, nwavels) array of wavelengths at which the spectrum is 
        defined. Input can also be a 1d
    weights : Array
        The (nspectra, nwavels) relative weights of each wavelength. Defaults 
        to uniform throughput.
    name : str
        The name of the spectrum.
    """


    def __init__(self        : Spectrum,
                 wavelengths : Array,
                 weights     : Array = None,
                 name        : str = 'CombinedSpectrum') -> Spectrum:
        """
        Constructor for the CombinedSpectrum class. Expects wavelengths and
        weights to have the same dimensionality, ie (nsources, nwavelengths).

        Parameters
        ----------
        wavelengths : Array, meters
            The (nspectra, nwavels) array of wavelengths at which the spectrum 
            is defined. Input can also be a 1d
        weights : Array (optional)
            The (nspectra, nwavels) relative weights of each wavelength. 
            Defaults to uniform throughput.
        name : str = 'CombinedSpectrum'
            The name of the spectrum.
        """
        super() # Access methods but don't instatiate attributes
        self.name = str(name)
        self.wavelengths = np.asarray(wavelengths, dtype=float)

        # Wavelengths
        # Tile single dimension wavelengths inputs
        if self.wavelengths.ndim == 1:
            self.wavelengths = np.tile(self.wavelengths, (2, 1))

        # Input checking
        assert self.wavelengths.ndim == 2, \
        ("Wavelengths must be a 2d array.")
        assert len(self.wavelengths[0]) > 0, \
        ("Length of wavelengths must be greater than 1.")

        # Weights
        weights = np.ones(self.wavelengths.shape)/self.wavelengths.shape[1] \
                                                if weights is None else weights
        self.weights = np.asarray(weights, dtype=float)

        # Tile single dimension weights inputs
        if self.weights.ndim == 1:
            self.weights = np.tile(self.weights, (len(self.wavelengths), 1))

        # Input checking
        assert self.weights.ndim == 2, \
        ("weights must be a 2d array.")
        assert len(self.weights[0]) > 0, \
        ("Length of weights must be greater than 1.")

        # Check consistency between wavelenghts and weights
        assert self.wavelengths.shape == self.weights.shape, "Weights and \
        Wavelengths must have the same shape."


    def normalise(self : Spectrum) -> Spectrum:
        """
        Method for returning a new spectrum object with a normalised total
        spectrum for each individual source.

        Returns
        -------
        spectrum : Specturm
            The spectrum object with the normalised spectrums.
        """
        weights = self.get_weights()
        total_power = weights.sum(1).reshape((len(weights), 1))
        norm_weights = weights/total_power
        return tree_at(lambda spectrum: spectrum.weights, self, norm_weights)
    

    def summary(self            : Spectrum, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        return (f"{self.name}: Applies a two spectrums with weights stored as "
                "arrays.")


from __future__ import annotations
import jax.numpy as np
from jax import vmap
from jax.tree_util import tree_map, tree_flatten
from equinox import tree_at
from zodiax import Base
from collections import OrderedDict
from copy import deepcopy
from inspect import signature
from typing import Union
import dLux


__all__ = ["model", "Instrument", "Optics", "Detector"]


Array = np.ndarray
Observation = lambda : dLux.observations.AbstractObservation
Source = lambda : dLux.sources.Source


###############
### Methods ###
###############
def model(optics      : Optics,
          sources     : Union[dict, list, Source],
          detector    : Detector = None,
          normalise   : bool     = True,
          flatten     : bool     = False,
          return_tree : bool     = False) -> Array:
    """
    A base level modelling function designed to robustly handle the different
    combinations of inputs. Models the sources through the instrument optics
    and detector. Users must provide optics and source.

    Parameters
    ----------
    optics : Optics
        The optics through which to model the source objects.
    sources : Union[dict, list, Source]
        The sources to observe.
    detector : Detector = None
        The detector to use with the observation.
    normalise : bool = None
        Whether to normalise the sources before modelling. Default is True.
    flatten : bool = False
        Whether the output image should be flattened. Default is False.
    return_tree : bool = False
        Whether to return a Pytree like object with matching tree structure as
        the input scene/sources/source. Default is False.

    Returns
    -------
    image : Array, Pytree
        The image of the scene modelled through the optics with detector and
        filter effects applied if they are supplied. Returns either as a single
        array (if return_tree is false), or a pytree like object with matching
        tree strucutre as the input scene/sources/source.
    """
    '''Input checking and formatting'''
    # Check that optics input is an Optics object.
    assert isinstance(optics, (Optics)), ("optics must be an Optics object.")

    # Check that detector input is a Detector object if specified.
    assert isinstance(detector, (Detector, type(None))), \
    ("detector must be a Detector object.")

    # Check that sources is a dict object.
    assert isinstance(sources, (dict, list, tuple, Source())), \
    ("sources must be a Source object, dict, list, or tuple object.")

    # Check that all inputs are Source objects
    if not isinstance(sources, Source()):
        source_vals = sources.values() if isinstance(sources, dict) else sources
        for source in source_vals:
            assert isinstance(source, Source()), \
            ("All entries within sources must be a Source object.")
    
    # Turn single source object into a list for mapping
    else:
        sources = [sources]

    # Normalise Sources
    if normalise:
        # Define the normalisation function
        normalise_fn = lambda source: source.normalise()

        # Map the normalisation function across the sources
        sources_in = tree_map(normalise_fn, sources, \
                is_leaf = lambda leaf: isinstance(leaf, Source()))
    else:
        sources_in = sources

        # Get sources
        sources_in = source.normalise() if normalise else source

    '''Begin modelling'''
    # Apply optional inputs
    model_fn = lambda source: source.model(optics)

    # Map the model_source function across the sources
    psf_tree = tree_map(model_fn, sources_in, 
            is_leaf = lambda leaf: isinstance(leaf, Source()))

    # Return psfs in the same structure as the sources
    if return_tree:

        # Apply detector if required
        if detector is not None:
            detector_fn = lambda psf: detector.apply_detector(psf)
            image_tree = tree_map(detector_fn, psf_tree, 
                is_leaf = lambda leaf: isinstance(leaf, np.ndarray))
        else:
            image_tree = psf_tree

        # flatten if required
        if flatten:
            flatten_fn = lambda image: image.flatten()
            tree_out = tree_map(flatten_fn, image_tree,
                is_leaf = lambda leaf: isinstance(leaf, np.ndarray))
        else:
            tree_out = image_tree

        # Return psfs with matching tree strucutre as input
        return tree_out


    # Return a single summed psf
    else:
        # Get flatten tree and sum to single psf
        psf = np.array(tree_flatten(psf_tree)[0]).sum(0)

        # Apply detector
        image = detector.apply_detector(psf) if detector is not None else psf

        # Flatten
        return image.flatten() if flatten else image


class Instrument(Base):
    """
    A high level class desgined to model the behaviour of a telescope. It
    stores a series different ∂Lux objects, and primarily passes the relevant
    information between these objects in order to coherently model some
    telescope observation.

    Attributes
    ----------
    optics : Optics
        A Optics object that defines some optical configuration.
    sources : dict
        A dictionary of the various source objects that the instrument is
        observing.
    detector : Detector
        A Detector object that is used to model the various
        instrumental effects on a psf.
    observation : Observation
        An class that inherits from Observation. This is to allow flexibility
        in the different kind of observations, ie applying dithers, switching
        filters, etc.
    """
    optics      : Optics
    sources     : dict
    detector    : Detector
    observation : Observation

    
    def __init__(self        : Instrument,
                 optics      : Optics,
                 sources     : Union[list, Source],
                 detector    : Detector    = None,
                 observation : Observation = None,
                 ) -> Instrument:
        """
        Constructor for the Instrument class.

        Parameters
        ----------
        optics : Optics
            A pre-configured Optics object.
        sources : Union[list, Source]
            Either a list of sources or an individual Source object.
        detector : Detector = None
            A pre-configured Detector object.
        observation : Observation = None
            An class that inherits from Observation. This is to allow
            flexibility in the different kind of observations, ie applying
            dithers, switching filters, etc.
        """
        # Optics
        if not isinstance(optics, Optics):
            raise ValueError("optics must be an Optics object.")
        self.optics = optics
        
        # Sources
        if isinstance(sources, Source()):
            sources = [sources]
        elif isinstance(sources, list):
            for source in sources:
                if not isinstance(source, Source()):
                    raise ValueError(
                        "sources must be a list of Source objects.")
        self.sources = dLux.utils.list_to_dictionary(sources, ordered=False)

        # Detector
        if not isinstance(detector, (Detector, type(None))):
            raise ValueError("detector must be an Detector object. "
                f"Got type {type(detector)}")
        self.detector = detector

        # Observation
        if not isinstance(observation, (Observation(), type(None))):
            raise ValueError("observation must be an Observation object.")
        self.observation = observation


    def observe(self : Instrument, **kwargs) -> Any:
        """
        Calls the `observe` method of the stored observation class, passing in
        any extra keyword arguments.

        Returns
        -------
         : Any
            The output of the stored observation class.
        """
        return self.observation.observe(self, **kwargs)


    def __getattr__(self : Instrument, key : str) -> object:
        """
        Magic method designed to allow accessing of the various items within
        the sub-dictionaries of this class via the 'class.attribute' method.
        It is recommended that each dictionary key in the optical layers,
        detector layers, and scene sources are unique to prevent unexpected
        behaviour. In the case they there are idenitcal keys across the
        dictionaries This method prioritises searching for keys in the optical
        layers, then detector layers, and then the scene sources.

        Parameters
        ----------
        key : str
            The key of the item to be searched for in the sub-dictionaries.

        Returns
        -------
        item : object
            The item corresponding to the supplied key in the sub-dictionaries.
        """
        if key in self.optics.layers.keys():
            return self.optics.layers[key]
        elif key in self.sources.keys():
            return self.sources[key]
        elif self.detector is not None and key in self.detector.layers.keys():
            return self.detector.layers[key]
        elif self.observation is not None and hasattr(self.observation, key):
            return getattr(self.observation, key)
        else:
            raise AttributeError("'{}' object has no attribute '{}'"\
                                 .format(type(self), key))


    def normalise(self : Instrument) -> Instrument:
        """
        Normalises the internally stored scene by calling the scene.normalise()
        method.

        Returns
        -------
        instrument : Instrument
            A new version of the instrument with the interally stored scene
            normalised.
        """
        leaf_fn = lambda source: isinstance(source, Source())
        normalise_fn = lambda source: source.normalise()
        return tree_at(lambda instrument: instrument.sources, self,
            tree_map(normalise_fn, self.sources, is_leaf=leaf_fn))
    

    def summarise(self : Instrument) -> None:
        """
        Prints a summary of all instrument
        """
        print("Sources summary:")
        summary_fn = lambda source: source.summarise()
        tree_map(summary_fn, self.sources)

        print("Optics summary:")
        self.optics.summarise()
        
        if self.detector is not None:
            print("Detector summary:")
            self.detector.summarise()


    def plot(self       : Optics, 
             wavelength : Array, 
             offset     : Array = np.zeros(2)) -> None:
        """
        Prints the summary of all the planes and then plots a wavefront as it
        propagates through the optics.

        Parameters
        ----------
        wavelength : Array, meters
            The wavelength of the wavefront to propagate through the optical
            layers.
        offset : Array, radians = np.zeros(2)
            The (x, y) offset from the optical axis of the source.
        """
        wf = self.optics.plot(wavelength, offset)
        self.detector.plot(wf.psf)


    def model(self              : Instrument,
              normalise_sources : bool = True,
              flatten           : bool = False,
              return_tree       : bool = False) -> Union(Array, dict):
        """
        A base level modelling function designed to robustly handle the
        different combinations of inputs. Models the sources through the
        instrument optics and detector.

        Parameters
        ----------
        normalise_sources : bool = True
            Whether to normalise the sources before modelling.
        flatten : bool = False
            Whether the output image should be flattened.
        return_tree : bool = False
            Whether to return a Pytree like object with matching tree structure
            as the input sources (ie dict).

        Returns
        -------
        image : Array, dict
            The image of the scene modelled through the optics with detector and
            filter effects applied if they are supplied. Returns either as a
            single array (if return_tree is false), or a dict of the output for
            each source.
        """
        return model(self.optics, self.sources, self.detector, 
            normalise_sources, flatten, return_tree)


class Optics(Base):
    """
    A high level class desgined to model the behaviour of some optical systems
    response to wavefronts.

    Attributes
    ----------
    layers: dict
        A collections.OrderedDict of 'layers' that define the transformations
        and operations upon some input wavefront through an optical system.
    """
    layers : OrderedDict


    def __init__(self : Optics, layers : list) -> Optics:
        """
        Constructor for the Optics class.

        Parameters
        ----------
        layers : list
            A list of ∂Lux 'layers' that define the transformations and
            operations upon some input wavefront through an optical system.
        """
        # Ensure input is a list
        assert isinstance(layers, list), ("Input layers must be a list, it is" \
        " automatically converted to a dictionary")

        # Ensure all entries are dLux layers
        for layer in layers:
            assert isinstance(layer, dLux.optics.OpticalLayer), ("All entries" \
            " within layers must be an dLux.optics.OpticalLayer object")

        self.layers = dLux.utils.list_to_dictionary(layers)


    def __getattr__(self : Optics, key : str) -> object:
        """
        Magic method designed to allow accessing of the various items within
        the layers dictionary of this class via the 'class.attribute' method.

        Parameters
        ----------
        key : str
            The key of the item to be searched for in the layers dictionary.

        Returns
        -------
        item : object
            The item corresponding to the supplied key in the layers dictionary.
        """
        if key in self.layers.keys():
            return self.layers[key]
        else:
            raise AttributeError("'{}' object has no attribute '{}'"\
                                 .format(type(self), key))


    def propagate_mono(self       : Optics,
                       wavelength : Array,
                       offset     : Array = np.zeros(2),
                       weight     : Array = np.array(1.)) -> Array:
        """
        Propagates a monochromatic point source through the optical layers.

        Parameters
        ----------
        wavelength : Array, meters
            The wavelength of the wavefront to propagate through the optical
            layers.
        offset : Array, radians, = np.zeros(2)
            The (x, y) offset from the optical axis of the source. Default
            value is (0, 0), on axis.
        weight : Array, = np.array(1.)
            The relative weighting of the wavelength. Simply scales the output
            psf.

        Returns
        -------
        psf : Array
            The monochromatic point spread function after being propagated
            though the optical layers.
        """
        # Ensure jax arrays
        wavelength = np.asarray(wavelength, dtype=float) \
                    if not isinstance(wavelength, np.ndarray) else wavelength
        offset = np.asarray(offset, dtype=float) \
                    if not isinstance(offset, np.ndarray) else offset
        weight = np.asarray(weight, dtype=float) \
                    if not isinstance(weight, np.ndarray) else weight

        # Ensure dimensionality
        assert wavelength.shape == (), "wavelength must be a scalar."
        assert offset.shape == (2,), "offset must be shape (2,), ie (x, y)."
        assert weight.shape == (), "weight must be a scalar."

        # Construct parameters dictionary
        params_dict = {"Wavefront"  : None,
                       "optics"     : self,
                       "wavelength" : wavelength,
                       "offset"     : offset}

        # Propagate though layers
        for key, layer in self.layers.items():
            params_dict = layer.apply(params_dict)
        psf = params_dict["Wavefront"].wavefront_to_psf()
        return weight * psf


    def propagate(self        : Optics,
                  wavelengths : Array,
                  offset      : Array = np.zeros(2),
                  weights     : Array = None) -> Array:
        """
        Propagates a broadband point source through the optical layers.

        Parameters
        ----------
        wavelengths : Array, meters
            The wavelengths of the wavefront to propagate through the optical
            layers.
        offset : Array, radians = np.zeros(2)
            The (x, y) offset from the optical axis of the source. Default
            value is (0, 0), on axis.
        weights : Array = None
            The relative weighting of the wavelengths. Simply scales the output
            psf.

        Returns
        -------
        psf : Array
            The broadband point spread function after being propagated
            though the optical layers.
        """
        # Format weights input
        wavelengths = np.asarray(wavelengths, dtype=float) \
                  if not isinstance(wavelengths, np.ndarray) else wavelengths
        assert wavelengths.ndim == 1, "wavelengths must be 1 dimensional.."

        # Format weights input
        if weights is None:
            weights = np.ones(len(wavelengths))/len(wavelengths)
        elif not isinstance(weights, np.ndarray):
            weights = np.asarray(weights, dtype=float)
        assert weights.ndim == 1, "weights must be 1 dimensional."

        # Ensure matching dimensionality
        assert wavelengths.shape == weights.shape, \
        ("wavelengths and weights must have the same shape.")

        # Offset checking
        offset = np.asarray(offset, dtype=float) \
                 if not isinstance(offset, np.ndarray) else offset
        assert offset.shape == (2,), "offset must be shape (2,), ie (x, y)."

        # Propagate
        propagator = vmap(self.propagate_mono, in_axes=(0, None, 0))
        psfs = propagator(wavelengths, offset, weights)
        return psfs.sum(0)


    def debug_prop(self       : Optics,
                   wavelength : Array,
                   offset     : Array = np.zeros(2),
                   weight     : Array = np.array(1.)) -> Array:
        """
        Propagates a monochromatic point source through the optical layers,
        while also returning the intermediate state of the parameter dictionary
        and layers after each layer application.

        Parameters
        ----------
        wavelength : Array, meters
            The wavelength of the wavefront to propagate through the optical
            layers.
        offset : Array, radians = np.zeros(2)
            The (x, y) offset from the optical axis of the source. Default
            value is (0, 0), on axis.
        weight : Array, = np.array(1.)
            The relative weighting of the wavelength. Simply scales the output
            psf.

        Returns
        -------
        psf : Array
            The monochromatic point spread function after being propagated
            though the optical layers.
        intermediate_dicts : list
            The intermediate states of the parameters dictionary.
        intermediate_layers : list
            The intermediate states of each layer after being applied to the
            wavefront.
        """
        # Ensure jax arrays
        wavelength = np.asarray(wavelength, dtype=float) \
                    if not isinstance(wavelength, np.ndarray) else wavelength
        offset = np.asarray(offset, dtype=float) \
                    if not isinstance(offset, np.ndarray) else offset
        weight = np.asarray(weight, dtype=float) \
                    if not isinstance(weight, np.ndarray) else weight

        # Ensure dimensionality
        assert wavelength.shape == (), "wavelength must be a scalar."
        assert offset.shape == (2,), "offset must be shape (2,), ie (x, y)."
        assert weight.shape == (), "weight must be a scalar."

        # Construct parameters dictionary
        params_dict = {"Wavefront"  : None,
                       "optics"     : self,
                       "wavelength" : wavelength,
                       "offset"     : offset}

        intermediate_dicts = []
        intermediate_layers = []
        for key, layer in self.layers.items():
            params_dict = layer.apply(params_dict)
            intermediate_dicts.append(params_dict.copy())
            intermediate_layers.append(deepcopy(layer))

        return params_dict["Wavefront"].wavefront_to_psf(), \
                                intermediate_dicts, intermediate_layers
    

    def get_planes(self : Optics) -> list:
        """
        Breaks the optical layers into planes, where each plane is a list of
        layers.

        Returns
        -------
        planes : list
            A list of lists, with the inner lists being optical layers, and the
            outer list being planes.
        """
        planes = []
        plane = []
        keys = self.layers.keys()
        for key in keys:
            layer = self.layers[key]
            plane.append(layer)
            if isinstance(layer, dLux.propagators.Propagator):
                planes.append(plane)
                plane = []
        return planes


    def summarise(self : Optics) -> None:
        """
        Prints a summary of all the planes in the optical system.
        """
        planes = self.get_planes()
        # TODO: Add plane type (Plane 0: Pupil)
        print("Text summary:")
        for i in range(len(planes)):
            print(f'Plane {i}')
            for layer in planes[i]:
                print(f"  {layer.summary(angular_units='arcseconds')}")
        print('\n')


    def plot(self       : Optics, 
             wavelength : Array, 
             offset     : Array = np.zeros(2)) -> None:
        """
        Prints the summary of all the planes and then plots a wavefront as it
        propagates through the optics.

        Parameters
        ----------
        wavelength : Array, meters
            The wavelength of the wavefront to propagate through the optical
            layers.
        offset : Array, radians = np.zeros(2)
            The (x, y) offset from the optical axis of the source.
        
        Returns
        -------
        wf : Wavefront
            The final wavefront after being propagated through the optical
            layers.
        """
        planes = self.get_planes()
        self.summarise()

        for i in range(len(planes)):
            print(f'Plane {i}')
            for layer in planes[i]:
                print(f"  {layer.summary()}")
                if isinstance(layer, dLux.CreateWavefront):
                    wf, parameters = layer(None, 
                                  {"wavelength": wavelength, 'offset': offset},returns_parameters=True)
                else:
                    # Inspect apply function to see if it takes/returns the parameters dict
                    input_parameters = signature(layer).parameters

                    # Method does not take in the parameters, update in place
                    if 'parameters' not in input_parameters:
                        wf = layer(wf)

                    # Method takes and return updated parameters
                    elif input_parameters['returns_parameters'].default == True:
                        wf, parameters = layer(wf, parameters)

                    # Method takes but does not return parameters
                    else:
                        wf = layer(wf, parameters)
                layer.display(wf)
        return wf


    def model(self              : Optics,
              sources           : Union[Source, dict, list],
              normalise_sources : bool = True,
              flatten           : bool = False,
              return_tree       : bool = False) -> Union(Array, dict):
        """
        A base level modelling function for modelling the optical system.
        Models the source or sources through the optics.

        Parameters
        ----------
        sources : Union[Source, dict, list]
            The source or sources to observe.
        normalise_sources : bool = True
            Whether to normalise the sources before modelling.
        flatten : bool = False
            Whether the output image should be flattened.
        return_tree : bool = False
            Whether to return a Pytree like object with matching tree structure
            as the input sources (ie dict).

        Returns
        -------
        image : Array, dict
            The image of the scene modelled through the optics. Returns either
            as a single array (if return_tree is false), or a dict of the output
            for each source.
        """
        # None input is for the detector
        return model(self, sources, None, normalise_sources, flatten,
            return_tree)


class Detector(Base):
    """
    A high level class desgined to model the behaviour of some detectors
    response to some psf.

    Attributes
    ----------
    layers: dict
        A collections.OrderedDict of 'layers' that define the transformations
        and operations upon some input psf as it interacts with the detector.
    """
    layers : OrderedDict


    def __init__(self : Detector, layers : list) -> Instrument:
        """
        Constructor for the Detector class.

        Parameters
        ----------
        layers : list
            An list of dLux detector layer classes that define the instrumental
            effects for some detector.
        """
        # Ensure input is a list
        assert isinstance(layers, list), ("Input layers must be a list, it is" \
        " automatically converted to a dictionary.")

        # Ensure all entries are dLux layers
        for layer in layers:
            assert isinstance(layer, dLux.detectors.DetectorLayer), \
            ("All entries within layers must be a "
             "dLux.detectors.DetectorLayer object.")

        # Construct layers
        self.layers = dLux.utils.list_to_dictionary(layers)


    def __getattr__(self : Detector, key : str) -> object:
        """
        Magic method designed to allow accessing of the various items within
        the layers dictionary of this class via the 'class.attribute' method.

        Parameters
        ----------
        key : str
            The key of the item to be searched for in the layers dictionary.

        Returns
        -------
        item : object
            The item corresponding to the supplied key in the layers dictionary.
        """
        if key in self.layers.keys():
            return self.layers[key]
        else:
            raise AttributeError("'{}' object has no attribute '{}'"\
                                 .format(type(self), key))


    def apply_detector(self : Instrument, image : Array) -> Array:
        """
        Applied the stored detector layers to the input image.

        Parameters
        ----------
        image : Array
            The input 'image' to the detector to be transformed.

        Returns
        -------
        image : Array
            The ouput 'image' after being transformed by the detector layers.
        """
        # Input type checking
        assert isinstance(image, np.ndarray), "Input must be a jax array."
        assert image.ndim == 2, "Input image must a 2d array."

        # Apply detector layers
        for key, layer in self.layers.items():
            image = layer(image)
        return image


    def debug_apply_detector(self  : Instrument, 
                             image : Array) -> Array:
        """
        Applied the stored detector layers to the input image, storing and
        returning the intermediate states of the image and layers.

        Parameters
        ----------
        image : Array
            The input 'image' to the detector to be transformed.

        Returns
        -------
        image : Array
            The ouput 'image' after being transformed by the detector layers.
        intermediate_images : list
            The intermediate states of the image.
        intermediate_layers : list
            The intermediate states of each layer after being applied to the
            image.
        """
        # Input type checking
        assert isinstance(image, np.ndarray), "Input must be a jax array."
        assert image.ndim == 2, "Input image must a 2d array."

        # Apply detector layers
        intermediate_images = []
        intermediate_layers = []
        for key, layer in self.layers.items():
            image = layer(image)
            intermediate_images.append(image.copy())
            intermediate_layers.append(deepcopy(layer))
        return image, intermediate_images, intermediate_layers
    

    def summarise(self : Detector) -> None:
        """
        Prints a summary of all the layers in the detector.
        """
        print("Text summary:")
        keys = self.layers.keys()
        for key in keys:
            layer = self.layers[key]
            print(f"  {layer.summary()}")
        print('\n')


    def plot(self : Optics, image : Array) -> None:
        """
        Prints the summary of all the layers and then plots a image as it
        propagates through the detector layer.

        Parameters
        ----------
        iamge : Array
            The image to propagate through the detector.
        """
        self.summarise()
        keys = self.layers.keys()
        for key in keys:
            layer = self.layers[key]
            print(f"{layer.summary()}")
            image = layer(image)
            layer.display(image)


    def model(self : Detector, image: Array) -> Array:
        """
        A function to apply the detector layers to the input image.

        Parameters
        ----------
        image: Array
            The image to be transformed by the detector layers.

        Returns
        -------
        image : Array
            The image after being transformed by the detector layers.
        """
        return self.apply_detector(image)


class Filter(Base):
    """
    NOTE: This class is under development.

    A class for modelling optical filters.

    Attributes
    ----------
    wavelengths : Array
        The wavelengths at which the filter is defined.
    throughput : Array
        The throughput of the filter at the corresponding wavelength.
    filter_name : str
        A string identifier that can be used to initialise specific filters.
    """
    wavelengths  : Array
    throughput   : Array
    filter_name  : str


    def __init__(self        : Filter,
                 wavelengths : Array = None,
                 throughput  : Array = None,
                 filter_name : str   = None) -> Filter:
        """
        Constructor for the Filter class. All inputs are optional and defaults
        to uniform unitary throughput. If filter_name is specified then
        wavelengths and weights must not be specified.

        Parameters
        ----------
        wavelengths : Array = None
            The wavelengths at which the filter is defined.
        throughput : Array = None
            The throughput of the filter at the corresponding wavelength.
        filter_name : str = None
            A string identifier that can be used to initialise specific filters.
            Currently no pre-built filters are implemented.
        """
        # Take the filter name as the priority input
        if filter_name is not None:
            # TODO: Pre load filters
            raise NotImplementedError("You know what this means.")
            pass

            # Check that wavelengths and throughput are not specified
            if wavelengths is not None or throughput is not None:
                raise ValueError("If filter_name is specified, wavelengths "
                "and throughput can not be specified.")

        # Check that both wavelengths and throughput are specified
        elif (wavelengths is     None and throughput is not None) or \
             (wavelengths is not None and throughput is     None):
            raise ValueError("If either wavelengths or throughput is "
            "specified, then both must be specified.")

        # Neither is specified
        elif wavelengths is None and throughput is None:
            self.wavelengths = np.array([0., np.inf])
            self.throughput  = np.array([1., 1.])
            self.filter_name = 'Unitary'

        # Both wavelengths and throughputs are specified
        else:
            self.wavelengths = np.asarray(wavelengths, dtype=float)
            self.throughput  = np.asarray(throughput,  dtype=float)
            self.filter_name = 'Custom'

            # Check bounds
            assert self.wavelengths.ndim == 1 and self.throughput.ndim == 1, \
            "Both wavelengths and throughput must be 1 dimensional."
            assert self.wavelengths.shape == self.throughput.shape, \
            ("wavelengths and throughput must have the same length.")
            assert np.min(self.wavelengths) >= 0, \
            ("wavelengths can not be less than 0.")
            assert (self.throughput >= 0).all() and \
            (self.throughput <= 1).all(), ("throughput must be between 0-1.")
            assert np.min(wavelengths) < np.max(wavelengths), \
            ("wavelengths must be in-order from small to large.")


    def get_throughput(self : Filter, sample_wavelenghts : Array) -> Array:
        """
        Gets the average throughput of the bandpass defined the the differences
        between each sample wavelength, ie if sample wavelengths are:
            [10, 20, 30, 40],
        the bandpasses for each sample wavelength will be
            [5-15, 15-25, 25-30, 35-40].
        The throughput is calculated as the average throughput over that
        bandpass.

        Parameters
        ----------
        sample_wavelengths : Array, meters
            The wavelengths at which to sample the filter. Must contain at
            least two values.

        Returns
        -------
        throughputs : Array
            The average throughput for each bandpass defined by
            sample_wavelengths.
        """
        mids = (sample_wavelenghts[1:] + sample_wavelenghts[:-1]) / 2
        diffs = np.diff(sample_wavelenghts)

        start = np.array([sample_wavelenghts[0] - diffs[0]/2])
        end = np.array([sample_wavelenghts[-1] + diffs[-1]/2])
        min_val = np.array([self.wavelengths.min()])
        max_val = np.array([self.wavelengths.max()])
        bounds = np.concatenate([start, mids, end])

        # Translate input wavelengths to indexes
        min_wavelength = self.wavelengths.min()
        max_wavelength = self.wavelengths.max()
        num_wavelength = len(self.wavelengths)
        wavelength_range = max_wavelength - min_wavelength
        bnd_indxs = num_wavelength * (bounds - min_wavelength)/wavelength_range
        bnd_indxs = np.clip(bnd_indxs, a_min=0, a_max=len(self.wavelengths))
        bnd_inds = np.round(bnd_indxs, decimals=0).astype(int)

        def nan_div(y, x):
            x_new = np.where(x == 0, 1, x)
            return np.where(x == 0, 0., y/x_new)

        def get_tp(start, end, weights, indexes):
            size = (end - start)
            val = np.where((indexes <= start) | (indexes >= end), \
                           0., weights).sum()
            return nan_div(val, size)

        starts = bnd_inds[:-1]
        ends   = bnd_inds[1:]
        # dwavelength = self.wavelengths[1] - self.wavelengths[0]
        indexes = np.arange(len(self.wavelengths))

        # weights = self.throughput/self.throughput.sum()
        weights = self.throughput
        out = vmap(get_tp, in_axes=(0, 0, None, None))(starts, ends, weights, indexes)
        return out


    def model(self : Filter, optics : Optics, **kwargs):
        """
        A base level modelling function designed to robustly handle the
        different combinations of inputs. Models the sources through the
        instrument optics and detector. Users must provide optics and some form
        of source, either via a scene, sources or single source input, but not
        multiple.

        Parameters
        ----------
        optics : Optics
            The optics through which to model the source objects.
        detector : Detector = None
            The detector to use with the observation. Defaults to the
            internally stored value.
        scene : Scene = None
            The scene to observe. Defaults to the internally stored value.
        sources : Union[dict, list, tuple) = None
            The sources to observe.
        source : dLux.sources.Source = None
            The source to observe.
        normalise_sources : bool = True
            Whether to normalise the sources before modelling. Default is True.
        flatten : bool = False
            Whether the output image should be flattened. Default is False.
        return_tree : bool = False
            Whether to return a Pytree like object with matching tree structure
            as the input scene/sources/source. Default is False.

        Returns
        -------
        image : Array, Pytree
            The image of the scene modelled through the optics with detector and
            filter effects applied if they are supplied. Returns either as a
            single array (if return_tree is false), or a pytree like object
            with matching tree strucutre as the input scene/sources/source.
        """
        return model(optics, filter=self, **kwargs)


import jax.numpy as np
from jax.tree_util import tree_map
from typing import Union
import dLux

__all__ = ["cartesian_to_polar", "polar_to_cartesian",
           "get_pixel_positions", "rotate", "translate", "shear", "compress"]


Array = np.ndarray


### Coordinates convertions ###
def cartesian_to_polar(coordinates : Array) -> Array:
    """
    Converts the input (x, y) cartesian cordinates into (r, phi) polar
    coordinates.

    Parameters
    ----------
    coordinates : Array
        The (x, y) cartesian coordinates to be converted into polar cordinates.

    Returns
    -------
    coordinates : Array
        The input cartesian coordinates converted into (r, phi) polar
        cordinates.
    """
    x, y = coordinates
    return np.array([np.hypot(x, y), np.arctan2(y, x)])


def polar_to_cartesian(coordinates : Array) -> Array:
    """
    Converts the input (r, phi) polar coordinates into (x, y) cartesian
    cordinates.

    Parameters
    ----------
    coordinates : Array
        The (r, phi) polar coordinates to be converted into cartesian
        cordinates.

    Returns
    -------
    coordinates : Array
        The input polar coordinates converted into (x, y) cartesian cordinates.
    """
    r, phi = coordinates
    return np.array([r*np.cos(phi), r*np.sin(phi)])


### Positions Calculations ###
def get_pixel_positions(npixels      : Union[int, tuple], 
                        pixel_scales : Union[tuple, float, None] = None,
                        offsets      : Union[tuple, float, None] = None,
                        polar        : bool = False,
                        indexing     : str = 'xy') -> Array:
    """
    Calculates the positions of the pixel centers for the given input. All 
    coordinates are output in units of meters. 
    
    The indexing argument is the same
    as in numpy.meshgrid., ie:  Giving the string ‘ij’ returns a meshgrid with 
    matrix indexing, while ‘xy’ returns a meshgrid with Cartesian indexing. In 
    the 2-D case with inputs of length M and N, the outputs are of shape (N, M) 
    for ‘xy’ indexing and (M, N) for ‘ij’ indexing. In the 3-D case with inputs 
    of length M, N and P, outputs are of shape (N, M, P) for ‘xy’ indexing and 
    (M, N, P) for ‘ij’ indexing. If the output is in polar coordainates, 
    indexing is set to 'xy' and the input must be 2d
    
    Parameters
    ----------
    npixels : Union[int, tuple]
        The number of pixels in each dimension.
    pixel_scales : Union[tuple, float, None] = None
        The pixel scales in each dimension. If a tuple, the length
        of the tuple must match the number of dimensions. If a float, the same
        scale is applied to all dimensions. If None, the scale is set to 1.
    offsets : Union[tuple, float, None] = None
        The offset of the pixel centers in each dimension. If a tuple, the 
        length of the tuple must match the number of dimensions. If a float, 
        the same offset is applied to all dimensions. If None, the offset is 
        set to 0.
    polar : bool = False
        If True, the output is in polar coordinates. If False, the output is in
        cartesian coordinates. Default is False.
    indexing : str = 'xy'
        The indexing of the output. Default is 'xy'. See numpy.meshgrid for more
        details.
    
    Returns
    -------
    positions : Array
        The positions of the pixel centers in the given dimensions.
    """
    if indexing not in ['xy', 'ij']:
        raise ValueError("indexing must be either 'xy' or 'ij'.")
    
    if polar and indexing == 'ij':
        indexing = 'xy'

    # Turn inputs into tuples
    if isinstance(npixels, int):
        npixels = (npixels,)

        if offsets is None:
            offsets = (0.,)
        elif not isinstance(offsets, (float, Array)):
            raise ValueError("offset must be a float or Array if npixels "
                             "is an int.")
        else:
            offsets = (offsets,)

        if pixel_scales is None:
            pixel_scales = (1.,)
        elif not isinstance(pixel_scales, (float, Array)):
            raise ValueError("pixel_scales must be a float or Array if npixels "
                             "is an int.")
        else:
            pixel_scales = (pixel_scales,)
        
    # Check input 
    else:
        if offsets is None:
            offsets = tuple([0.]*len(npixels))
        elif not isinstance(offsets, tuple):
            raise ValueError("offset must be an be a float or Array if npixels "
                             "is an int.")
        else:
            if len(offsets) != len(npixels):
                raise ValueError("offset must have the same length as npixels.")
            
        if pixel_scales is None:
            pixel_scales = tuple([1.]*len(npixels))
        elif isinstance(pixel_scales, float):
            pixel_scales = tuple([pixel_scales]*len(npixels))
        elif not isinstance(pixel_scales, tuple):
            raise ValueError("pixel_scales must be a tuple if npixels is a tuple.")
        else:
            if len(pixel_scales) != len(npixels):
                raise ValueError("pixel_scales must have the same length as npixels.")
    
    def pixel_fn(n, offset, scale):
        pix = np.arange(n) - (n - 1) / 2.
        pix *= scale
        pix -= offset
        return pix
    
    pixels = tree_map(pixel_fn, npixels, offsets, pixel_scales)

    # ouput (x, y) for 2d, else in order
    positions = np.array(np.meshgrid(*pixels, indexing=indexing))

    if polar:
        if len(npixels) != 2:
            raise ValueError("polar coordinates are only defined for 2D arrays.")
        return cartesian_to_polar(positions)

    # Squeeze for empty axis removal with 1d
    return np.squeeze(positions)


### Coordinates Transformations ###
def rotate(coordinates: Array, rotation: Array) -> Array:
    """
    Rotate the coordinate system by a pre-specified amount.

    Parameters
    ----------
    coordinates : Array, meters
        A `(2, npix, npix)` representation of the coordinate 
        system. The leading dimensions specifies the x and then 
        the y coordinates in that order. 
    rotation : Array, radians
        The counter-clockwise rotation to apply.

    Returns
    -------
    coordinates : Array, meters
        The rotated coordinate system. 
    """
    x, y = coordinates[0], coordinates[1]
    new_x = np.cos(-rotation) * x + np.sin(-rotation) * y
    new_y = -np.sin(-rotation) * x + np.cos(-rotation) * y
    return np.array([new_x, new_y])


def translate(coordinates: Array, centre: Array) -> Array:
    """
    Move the center of the coordinate system by some 
    amount (centre). 

    Parameters
    ----------
    coordinates : Array, meters
        The (x, y) coordinates with the dimensions 
        (2, npix, npix).
    centre : Array, meters
        The (x, y) coordinates of the new centre 
        with dimensions (2,)

    Returns
    -------
    coordinates: Array, meters
        The translated coordinate system. 
    """
    return coordinates - centre[:, None, None]


def shear(coordinates: Array, shear: Array) -> Array:
    """
    Apply a shear to the coordinate system. 

    Parameters
    ----------
    coordinates : Array, meters
        The (x, y) coordinates with the dimensions 
        (2, npix, npix).
    shear : Array
        The (x, y) shear with dimensions (2,)

    Returns
    -------
    coordinates: Array, meters
        The sheared coordinate system. 
    """
    trans_coordinates: Array = np.transpose(coordinates, (0, 2, 1))
    return coordinates + trans_coordinates * shear[:, None, None]


def compress(coordinates: Array, compression: Array) -> Array:
    """
    Apply a compression to the coordinates.

    Parameters
    ----------
    coordinates : Array, meters
        The (x, y) coordinates with the dimensions 
        (2, npix, npix).
    compression : Array
        The (x, y) compression with dimensions (2,)

    Returns
    -------
    coordinates : Array, meters
        The compressed coordinates. 
    """
    return coordinates * compression[:, None, None]





name = "dLux"
__version__ = "0.12.0"

# Import as modules
from . import core
from . import detectors
from . import optics
from . import wavefronts
from . import propagators
from . import utils
from . import apertures
from . import aberrations
from . import observations
from . import sources
from . import spectrums
from . import exceptions

# Import core functions from modules
from .core         import *
from .detectors    import *
from .optics       import *
from .wavefronts   import *
from .propagators  import *
from .apertures    import *
from .aberrations  import *
from .observations import *
from .sources      import *
from .spectrums    import *
from .exceptions   import *

# Add to __all__
__all__ = core.__all__ + detectors.__all__ + optics.__all__ + \
    wavefronts.__all__ + propagators.__all__ + sources.__all__ + \
    spectrums.__all__ + apertures.__all__ + aberrations.__all__ + \
    observations.__all__

# Check for 64-bit
from jax import config
if not config.x64_enabled:
    print("dLux: Jax is running in 32-bit, to enable 64-bit visit: "
          "https://jax.readthedocs.io/en/latest/notebooks/"
          "Common_Gotchas_in_JAX.html#double-64bit-precision")


from __future__ import annotations
import jax.numpy as np
import pytest
import dLux
from jax import config
config.update("jax_debug_nans", True)


class TestCartesianMFT():
    """
    Test the CartesianMFT class.
    """


    def test_constructor(self, create_cartesian_mft : callable):
        """
        Tests the constructor.
        """
        # Test constructor
        create_cartesian_mft()


    def test_propagate(self, create_cartesian_mft : callable):
        """
        Tests the propagate method.
        """
        wl, npix, f_pscale, fl = 1e-6, 32, 1e-6, 5.
        p_pscale = 1/npix
        amplitude = np.ones((1, npix, npix))
        phase = np.zeros((1, npix, npix))
        plane_type = dLux.PlaneType.Pupil

        # Construct
        wf = dLux.CartesianWavefront(wl, p_pscale, amplitude, phase, plane_type)
        prop = create_cartesian_mft(npix, f_pscale, fl)
        inv_prop = create_cartesian_mft(npix, p_pscale, fl, inverse=True)

        # Prop
        focal = prop(wf)
        pupil = inv_prop(focal)

        # Test
        assert not np.isnan(focal.psf).any()
        assert not np.isnan(pupil.psf).any()

        # Shift tests
        shift_pix = np.ones(2)
        shift = f_pscale * np.ones(2)

        # Construct
        prop_shift = create_cartesian_mft(npix, f_pscale, fl, shift=shift)
        prop_shift_pix = create_cartesian_mft(npix, f_pscale, fl, 
                                            shift=shift_pix, pixel_shift=True)

        # Prop
        focal_shift = prop_shift(wf)
        focal_shift_pix = prop_shift_pix(wf)
        focal_roll = np.roll(focal.psf, (1, 1), (0, 1))[1:, 1:]

        # Test
        assert np.allclose(focal_roll, focal_shift.psf[1:, 1:])
        assert np.allclose(focal_roll, focal_shift_pix.psf[1:, 1:])


class TestAngularMFT():
    """
    Test the AngularMFT class.
    """

    def test_constructor(self, create_angular_mft : callable):
        """
        Tests the constructor.
        """
        # Test constructor
        create_angular_mft()


    def test_propagate(self, create_angular_mft : callable):
        """
        Tests the propagate method.
        """
        wl, npix, f_pscale = 1e-6, 32, 2e-7
        p_pscale = 1/npix
        amplitude = np.ones((1, npix, npix))
        phase = np.zeros((1, npix, npix))
        plane_type = dLux.PlaneType.Pupil

        # Construct
        wf = dLux.AngularWavefront(wl, p_pscale, amplitude, phase, plane_type)
        prop = create_angular_mft(npix, f_pscale)
        inv_prop = create_angular_mft(npix, p_pscale, inverse=True)

        # Prop
        focal = prop(wf)
        pupil = inv_prop(focal)

        # Test
        assert not np.isnan(focal.psf).any()
        assert not np.isnan(pupil.psf).any()

        # Shift tests
        shift_pix = np.ones(2)
        shift = f_pscale * np.ones(2)

        # Construct
        prop_shift = create_angular_mft(npix, f_pscale, shift=shift)
        prop_shift_pix = create_angular_mft(npix, f_pscale, 
                                            shift=shift_pix, pixel_shift=True)

        # Prop
        focal_shift = prop_shift(wf)
        focal_shift_pix = prop_shift_pix(wf)
        focal_roll = np.roll(focal.psf, (1, 1), (0, 1))[1:, 1:]

        # Test
        assert np.allclose(focal_roll, focal_shift.psf[1:, 1:])
        assert np.allclose(focal_roll, focal_shift_pix.psf[1:, 1:])


class TestCartesianFFT():
    """
    Test the CartesianFFT class.
    """


    def test_constructor(self, create_cartesian_fft : callable):
        """
        Tests the constructor.
        """
        # Test constructor
        create_cartesian_fft()


    def test_propagate(self, create_cartesian_fft : callable):
        """
        Tests the propagate method.
        """
        wl, npix, fl = 1e-6, 32, 5.
        p_pscale = 1/npix
        amplitude = np.ones((1, npix, npix))
        phase = np.zeros((1, npix, npix))
        plane_type = dLux.PlaneType.Pupil

        # Construct
        wf = dLux.CartesianWavefront(wl, p_pscale, amplitude, phase, plane_type)
        prop = create_cartesian_fft(fl)
        inv_prop = create_cartesian_fft(fl, inverse=True)

        # Prop
        focal = prop(wf.pad_to(npix * 5))
        pupil = inv_prop(focal).crop_to(npix)

        # Test
        assert not np.isnan(focal.psf).any()
        assert not np.isnan(pupil.psf).any()

class TestAngularFFT():
    """
    Test the AngularFFT class.
    """


    def test_constructor(self, create_angular_fft : callable):
        """
        Tests the constructor.
        """
        # Test constructor
        create_angular_fft()


    def test_propagate(self, create_angular_fft : callable):
        """
        Tests the propagate method.
        """
        wl, npix = 1e-6, 32
        p_pscale = 1/npix
        amplitude = np.ones((1, npix, npix))
        phase = np.zeros((1, npix, npix))
        plane_type = dLux.PlaneType.Pupil

        # Construct
        wf = dLux.AngularWavefront(wl, p_pscale, amplitude, phase, plane_type)
        prop = create_angular_fft()
        inv_prop = create_angular_fft(inverse=True)

        # Prop
        focal = prop(wf.pad_to(npix * 5))
        pupil = inv_prop(focal).crop_to(npix)

        # Test
        assert not np.isnan(focal.psf).any()
        assert not np.isnan(pupil.psf).any()



class TestCartesianFresnel():
    """
    Test the CartesianFresnel class.
    """


    def test_constructor(self, create_cartesian_fresnel : callable):
        """
        Tests the constructor.
        """
        # Test constructor
        create_cartesian_fresnel()


    def test_propagate(self, create_cartesian_fresnel : callable):
        """
        Tests the propagate method.
        """
        wl, npix, f_pscale, fl = 1e-6, 32, 2e-7, 5.
        p_pscale = 1/npix
        amplitude = np.ones((1, npix, npix))
        phase = np.zeros((1, npix, npix))
        plane_type = dLux.PlaneType.Pupil

        # Construct
        wf = dLux.FarFieldFresnelWavefront(wl, p_pscale, amplitude, phase, 
                                         plane_type)
        fresnel_prop = create_cartesian_fresnel(npix, f_pscale, fl, 5e-5)
        focal_prop = dLux.CartesianFresnel(npix, f_pscale, fl, 0.)
        inv_prop = create_cartesian_fresnel(npix, p_pscale, fl, 1e0, inverse=True)

        # Prop
        fresnel = fresnel_prop(wf)
        focal = focal_prop(wf)
        pupil = inv_prop(focal)

        # Test
        assert not np.isnan(fresnel.psf).any()
        assert not np.isnan(focal.psf).any()
        assert not np.isnan(pupil.psf).any()



from __future__ import annotations
import jax.numpy as np
from equinox import tree_at
from abc import ABC, abstractmethod
from dLux.utils.coordinates import get_pixel_positions
from dLux.utils.helpers import two_image_plot
from dLux.utils.units import convert_angular, convert_cartesian
import dLux


__all__ = ["CartesianMFT", "AngularMFT", "CartesianFFT", "AngularFFT",
           "CartesianFresnel"]


Array = np.ndarray


########################
### Abstract Classes ###
########################
class Propagator(dLux.optics.OpticalLayer, ABC):
    """
    An abstract class to store the various properties of the propagation of
    some wavefront.

    Attributes
    ----------
    inverse : bool
        Is this an 'inverse' propagation. Non-inverse propagations represents
        propagation from a pupil to a focal plane, and inverse represents
        propagation from a focal to a pupil plane.
    name : str
        The name for this propagator.
    """
    inverse : bool


    def __init__(self    : Propagator,
                 inverse : bool = False,
                 **kwargs) -> Propagator:
        """
        Constructor for the Propagator.

        Parameters
        ----------
        inverse : bool = False
            Is this an 'inverse' propagation. Non-inverse propagations
            represents propagation from a pupil to a focal plane, and inverse
            represents propagation from a focal to a pupil plane.
        """
        super().__init__(**kwargs)
        assert isinstance(inverse, bool), ("inverse must be a boolean.")
        self.inverse = bool(inverse)


    @abstractmethod
    def propagate(self : Propagator, wavefront : Wavefront) -> Array: # pragma: no cover
        """
        Performs the propagation as a directional wrapper to the fourier methods
        of the class.

        Parameters
        ----------
        wavefront : Wavefront
            The `Wavefront to propagate.

        Returns
        -------
        field : Array
            The normalised electric field of the wavefront after propagation.
        """
        return


class VariableSamplingPropagator(Propagator, ABC):
    """
    A propagator that implements the Soummer et. al. 2007 MFT algorithm
    allowing variable sampling in the outuput plane rather than the fixed
    sampling enforced by Fast Fourier Transforms(FFTs).

    Attributes
    ----------
    npixels_out : int
        The number of pixels in the output plane.
    pixel_scale_out : Array, meters/pixel or radians/pixel
        The pixel scale in the output plane, measured in meters or radians per
        pixel for Cartesian or Angular Wavefront respectively.
    shift : Array
        The (x, y) shift to apply to the wavefront in the output plane.
    pixel_shift : bool
        Should the shift value be considered in units of pixels, or in the
        physical units of the output plane (ie pixels or meters, radians). True
        interprets the shift value in pixel units.
    inverse : bool
        Is this an 'inverse' propagation. Non-inverse propagations represents
        propagation from a pupil to a focal plane, and inverse represents
        propagation from a focal to a pupil plane.
    name : str
        The name for this propagator.
    """
    npixels_out     : int
    pixel_scale_out : Array
    shift           : Array
    pixel_shift     : bool

    def __init__(self            : Propagator,
                 pixel_scale_out : Array,
                 npixels_out     : int,
                 shift           : Array = np.array([0., 0.]),
                 pixel_shift     : bool  = False,
                 **kwargs) -> Propagator:
        """
        Constructor for VariableSampling propagators.

        Parameters
        ----------
        npixels_out : int
            The number of pixels in the output plane.
        pixel_scale_out : Array, meters/pixel or radians/pixel
            The pixel scale in the output plane, measured in meters or radians
            per pixel for Cartesian or Angular Wavefront respectively.
        shift : Array = np.array([0., 0.])
            The (x, y) shift to apply to the wavefront in the output plane.
        pixel_shift : bool = False
            Should the shift value be considered in units of pixel, or in the
            physical units of the output plane (ie pixels or meters, radians). =
            True interprets the shift value in pixel units.
        """
        super().__init__(**kwargs)
        self.pixel_scale_out = np.asarray(pixel_scale_out, dtype=float)
        self.npixels_out     = int(npixels_out)
        self.shift           = np.asarray(shift, dtype=float)
        self.pixel_shift     = bool(pixel_shift)
        assert self.pixel_scale_out.ndim == 0, \
        ("pixel_scale_out must be a scalar.")
        assert self.shift.shape == (2,), \
        ("shift must be an array of shape (2,) ie (x, y).")


    @abstractmethod
    def get_nfringes(self : Propagator, wavefront : Wavefront) -> Array: # pragma: no cover
        """
        The number of diffraction fringes in the output plane.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront being propagated.

        Returns
        -------
        fringes : Array
            The number of diffraction fringes in the output plane.
        """
        return


    def get_shift(self : Propagator) -> Array:
        """
        Accessor for the shift parameter. Converts to units of pixels if the
        pixel_shift parameter is True.

        Returns
        -------
        shift : Array
            The (x, y) shift to apply to the wavefront throughout the
            propagation.
        """
        return self.shift if self.pixel_shift else \
               self.shift / self.pixel_scale_out


    def _generate_transfer_matrices(self         : Propagator,
                                    pixel_offset : Array,
                                    pixel_scales : tuple,
                                    npixels      : tuple) -> Array:
        """
        The transfer matrices for the fourier transforms.

        Parameters
        ----------
        pixel_offset : Array, pixels
            The offset in units of pixels.
        pixel_scales : tuple
            The pixel_scale values at the input and output planes respectively.
        npixels : tuple
            The number of pixels at the input and output planes respectively.

        Returns
        -------
        transfer_matrices : Array
            The transfer matrices.
        """
        input_scale, output_scale = pixel_scales
        pixels_input, npixels_out = npixels
        sign = 1 if self.inverse else -1

        input_coordinates = get_pixel_positions(pixels_input, input_scale,
                                                    pixel_offset * input_scale)

        output_coordinates = get_pixel_positions(npixels_out, output_scale,
                                                    pixel_offset * output_scale)

        input_to_output = np.outer(input_coordinates, output_coordinates)

        return np.exp(-2. * sign * np.pi * 1j * input_to_output)


    def propagate(self      : Propagator,
                  wavefront : Wavefront) -> Array:
        """
        Propagates the wavefront from the input plane to the output plane using
        a Matrix Fourier Transform.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront being propagated.

        Returns
        -------
        field : Array
            The normalised electric field phasor after the propagation.
        """
        field = wavefront.phasor
        nfields = wavefront.nfields

        input_scale = 1.0 / wavefront.npixels
        output_scale = self.get_nfringes(wavefront) / self.npixels_out
        npixels_in = wavefront.npixels
        npixels_out = self.npixels_out
        x_offset, y_offset = self.get_shift()

        x_matrix = np.tile(self._generate_transfer_matrices(
                    x_offset, (input_scale, output_scale),
                    (npixels_in, npixels_out)), (nfields, 1, 1))

        y_matrix = np.tile(self._generate_transfer_matrices(
                    y_offset, (input_scale, output_scale),
                    (npixels_in, npixels_out)).T, (nfields, 1, 1))

        output_field = (y_matrix @ field) @ x_matrix

        normalising_factor = np.exp(np.log(self.get_nfringes(wavefront)) - \
                            (np.log(npixels_in) + np.log(npixels_out)))

        return output_field * normalising_factor


    def __call__(self : Propagator, wavefront : Wavefront) -> Wavefront:
        """
        Propagates the `Wavefront`.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        wavefront : Wavefront
            The wavefront with the optical layer applied.
        """
        phasor = self.propagate(wavefront)

        new_amplitude  = np.abs(phasor)
        new_phase      = np.angle(phasor)
        new_plane_type = dLux.PlaneType.Pupil if self.inverse else \
                         dLux.PlaneType.Focal

        return tree_at(lambda wavefront: \
                       (wavefront.amplitude,
                        wavefront.phase,
                        wavefront.plane_type,
                        wavefront.pixel_scale),
                        wavefront,
                       (new_amplitude,
                        new_phase,
                        new_plane_type,
                        self.pixel_scale_out))


class FixedSamplingPropagator(Propagator, ABC):
    """
    A propagator that implements the Fast Fourier Transform algorithm. This
    algorith has a fixed sampling in the output plane, at one fringe per pixel.
    Note the size of the 'fringe' in this context is similar to an optical
    fringe in that its angular size is calcualted via wavelength/wavefront
    diameter.

    These propagators are implemented using the jax.numpy.fft package, with the
    appropriate normalisations and pixel sizes tracked for optical propagation.

    Attributes
    ----------
    inverse : bool
        Is this an 'inverse' propagation. Non-inverse propagations represents
        propagation from a pupil to a focal plane, and inverse represents
        propagation from a focal to a pupil plane.
    name : str
        The name for this propagator.
    """


    def __init__(self : Propagator, **kwargs) -> Propagator:
        """
        Constructor for FixedSampling propagators.
        """
        super().__init__(**kwargs)


    @abstractmethod
    def get_pixel_scale_out(self     : Propagator,
                           wavefront : Wavefront) -> Array: # pragma: no cover
        """
        Calculates the pixel scale in the output plane.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront that is being propagated.

        Returns
        -------
        pixel_scale : Array, meters/pixel or radians/pixel
            The pixel scale in the output plane, measured in meters or radians
            per pixel for Cartesian or Angular Wavefront respectively.
        """
        return


    def propagate(self       : Propagator,
                   wavefront : Wavefront) -> Array:
        """
        Propagates the wavefront by perfroming a Fast Fourier Transform.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        field : Array
            The normalised electric field phasor after the propagation.
        """
        if self.inverse:
            output_field = np.fft.fft2(np.fft.ifftshift(wavefront.phasor))
        else:
            output_field = np.fft.fftshift(np.fft.ifft2(wavefront.phasor))

        normalising_factor = self.inverse / wavefront.npixels + \
                             (1 - self.inverse) * wavefront.npixels

        return output_field * normalising_factor


    def __call__(self : Propagator, wavefront : Wavefront) -> Wavefront:
        """
        Propagates the `Wavefront`.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        wavefront : Wavefront
            The wavefront with the optical layer applied.
        """
        phasor = self.propagate(wavefront)

        new_amplitude  = np.abs(phasor)
        new_phase      = np.angle(phasor)
        new_plane_type = dLux.PlaneType.Pupil if self.inverse else \
                         dLux.PlaneType.Focal

        return tree_at(lambda wavefront: \
                       (wavefront.amplitude,
                        wavefront.phase,
                        wavefront.plane_type,
                        wavefront.pixel_scale),
                        wavefront,
                       (new_amplitude,
                        new_phase,
                        new_plane_type,
                        self.get_pixel_scale_out(wavefront)))


class CartesianPropagator(Propagator, ABC):
    """
    A propagator class to store the focal_length parameter for cartesian
    propagations defined by a physical propagation distance defined as
    focal_length.

    Attributes
    ----------
    focal_length : Array, meters
        The focal_length of the lens/mirror this propagator represents.
    inverse : bool
        Is this an 'inverse' propagation. Non-inverse propagations represents
        propagation from a pupil to a focal plane, and inverse represents
        propagation from a focal to a pupil plane.
    name : str
        The name for this propagator.
    """
    focal_length : Array


    def __init__(self         : Propagator,
                 focal_length : Array,
                 **kwargs) -> Propagator:
        """
        Constructor for Cartesian propagators.

        Parameters
        ----------
        focal_length : Array, meters
            The focal_length of the lens/mirror this propagator represents.
        """
        super().__init__(**kwargs)
        self.focal_length = np.asarray(focal_length, dtype=float)
        assert self.focal_length.ndim == 0, ("focal_length must a scalar.")


class AngularPropagator(Propagator, ABC):
    """
    A simple propagator class designed to be inhereited by propagators that
    operate on wavefronts defined in angular units in focal planes.

    Attributes
    ----------
    inverse : bool
        Is this an 'inverse' propagation. Non-inverse propagations represents
        propagation from a pupil to a focal plane, and inverse represents
        propagation from a focal to a pupil plane.
    name : str
        The name for this propagator.
    """


    def __init__(self : Propagator, **kwargs) -> Propagator:
        """
        Constructor for Angular propagators.
        """
        super().__init__(**kwargs)


class FarFieldFresnel(Propagator, ABC):
    """
    A propagator class to store the propagation_shift parameter required for
    Far-Field fresnel propagations. These classes implement algorithms that use
    quadratic phase factors to better represent out-of-plane behaviour of
    wavefronts, close to the focal plane.

    Attributes
    ----------
    propagation_shift : Array, meters
        The shift in the propagation distance of the wavefront.
    inverse : bool
        Is this an 'inverse' propagation. Non-inverse propagations represents
        propagation from a pupil to a focal plane, and inverse represents
        propagation from a focal to a pupil plane.
    name : str
        The name for this propagator.
    """
    propagation_shift : Array


    def __init__(self, propagation_shift, **kwargs) -> Propagator:
        """
        Constructor for FarFieldFresnel propagators.

        Parameters
        ----------
        propagation_shift : Array, meters
            The shift in the propagation distance of the wavefront.
        """
        super().__init__(**kwargs)
        self.propagation_shift  = np.asarray(propagation_shift,  dtype=float)
        assert self.propagation_shift.ndim == 0, \
        ("propagation_shift must be scalar array.")


########################
### Concrete Classes ###
########################
class CartesianMFT(CartesianPropagator, VariableSamplingPropagator):
    """
    A Propagator class designed to propagate a wavefront to a plane that is
    defined in cartesian units (ie meters/pixel), with a variable output
    sampling in that plane.

    Attributes
    ----------
    npixels_out : int
        The number of pixels in the output plane.
    pixel_scale_out : Array, meters/pixel
        The pixel scale in the output plane, measured in meters per pixel.
    focal_length : Array, meters
        The focal_length of the lens/mirror this propagator represents.
    inverse : bool
        Is this an 'inverse' propagation. Non-inverse propagations represents
        propagation from a pupil to a focal plane, and inverse represents
        propagation from a focal to a pupil plane.
    shift : Array
        The (x, y) shift to apply to the wavefront in the output plane.
    pixel_shift : bool
        Should the shift value be considered in units of pixels, or in the
        physical units of the output plane (ie pixels or meters, radians). True
        interprets the shift value in pixel units.
    name : str
        The name for this propagator.
    """


    def __init__(self            : Propagator,
                 npixels_out     : int,
                 pixel_scale_out : Array,
                 focal_length    : Array,
                 inverse         : bool  = False,
                 shift           : Array = np.array([0., 0.]),
                 pixel_shift     : bool  = False,
                 name            : str   = 'CartesianMFT') -> Propagator:
        """
        Parameters
        ----------
        npixels_out : int
            The number of pixels in the output plane.
        pixel_scale_out : Array, meters/pixel
            The pixel scale in the output plane, measured in meters per pixel.
        focal_length : Array, meters
            The focal_length of the lens/mirror this propagator represents.
        inverse : bool = False
            Is this an 'inverse' propagation. Non-inverse propagations
            represents propagation from a pupil to a focal plane, and inverse
            represents propagation from a focal to a pupil plane.
        shift : Array = np.array([0., 0.])
            The (x, y) shift to apply to the wavefront in the output plane.
        pixel_shift : bool = False
            Should the shift value be considered in units of pixel, or in the
            physical units of the output plane (ie pixels or meters, radians).
        name : str = 'CartesianMFT'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name            = name,
                         inverse         = inverse,
                         shift           = shift,
                         pixel_shift     = pixel_shift,
                         focal_length    = focal_length,
                         pixel_scale_out = pixel_scale_out,
                         npixels_out     = npixels_out)


    def get_nfringes(self      : Propagator,
                     wavefront : Wavefront) -> Array:
        """
        The number of diffraction fringes in the output plane.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront being propagated.

        Returns
        -------
        fringes : Array
            The number of diffraction fringes in the output plane.
        """
        size_in = wavefront.diameter
        size_out = self.pixel_scale_out * self.npixels_out
        return size_in * size_out / self.focal_length / wavefront.wavelength
    

    def summary(self            : Propagator, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        plane_out = 'pupil' if self.inverse else 'focal'

        pixel_scale = convert_cartesian(self.pixel_scale_out, 'meters', 
                                        cartesian_units)
        focal_length = convert_cartesian(self.focal_length, 'meters', 
                                        cartesian_units)
        shift = convert_cartesian(self.shift, 'meters', cartesian_units) \
            if not self.pixel_shift else self.shift

        prop_string = (f"Propagates the Wavefront {focal_length} "
                       f"{cartesian_units} to a {plane_out} plane "
                       f"with {self.npixels_out}x{self.npixels_out} pixels of "
                       f"size {pixel_scale:.{sigfigs}} {cartesian_units} ")
        
        if (self.shift == np.zeros(2)).all():
            return prop_string +  "using a 2-sided MFT."
        else:
            if self.pixel_shift:
                shift_string = (f"shifted by {shift} pixels in the (x, y) "
                                "dimension ")
            else:
                shift_string = (f"shifted by {shift} {cartesian_units} in the "
                                "(x, y) dimension ")
            return prop_string + shift_string + "using a 2-sided MFT."
        

class AngularMFT(AngularPropagator, VariableSamplingPropagator):
    """
    A Propagator class designed to propagate wavefronts, with pixel scale units
    defined in meters per pixel in pupil planes and radians/pixel in focal
    planes, with a variable output sampling in the output plane.

    Attributes
    ----------
    npixels_out : int
        The number of pixels in the output plane.
    pixel_scale_out : Array, meters/pixel or radians/pixel
        The pixel scale in the output plane, measured in meters per pixel in
        pupil plane and radians per pixel in focal planes.
    inverse : bool
        Is this an 'inverse' propagation. Non-inverse propagations represents
        propagation from a pupil to a focal plane, and inverse represents
        propagation from a focal to a pupil plane.
    shift : Array
        The (x, y) shift to apply to the wavefront in the output plane.
    pixel_shift : bool
        Should the shift value be considered in units of pixels, or in the
        physical units of the output plane (ie pixels or meters, radians). True
        interprets the shift value in pixel units.
    name : str
        The name for this propagator.
    """


    def __init__(self            : Propagator,
                 npixels_out     : int,
                 pixel_scale_out : Array,
                 inverse         : bool  = False,
                 shift           : Array = np.array([0., 0.]),
                 pixel_shift     : bool  = False,
                 name            : str   = 'AngularMFT') -> Propagator:
        """
        Parameters
        ----------
        npixels_out : int
            The number of pixels in the output plane.
        pixel_scale_out : Array, radians/pixel, meters/pixel
            The pixel scale in the output plane, measured in meters per pixel in
            pupil planes and radians per pixel in focal planes.
        inverse : bool = False
            Is this an 'inverse' propagation. Non-inverse propagations
            represents propagation from a pupil to a focal plane, and inverse
            represents propagation from a focal to a pupil plane.
        shift : Array = np.array([0., 0.])
            The (x, y) shift to apply to the wavefront in the output plane.
        pixel_shift : bool = False
            Should the shift value be considered in units of pixel, or in the
            physical units of the output plane (ie pixels or meters, radians).
        name : str = 'AngularMFT'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name            = name,
                         inverse         = inverse,
                         shift           = shift,
                         pixel_shift     = pixel_shift,
                         pixel_scale_out = pixel_scale_out,
                         npixels_out     = npixels_out)


    def get_nfringes(self      : Propagator,
                     wavefront : Wavefront) -> Array:
        """
        The number of diffraction fringes in the output plane.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront being propagated.

        Returns
        -------
        fringes : Array
            The number of diffraction fringes in the output plane.
        """
        fringe_size = wavefront.wavelength / wavefront.diameter
        detector_size = self.npixels_out * self.pixel_scale_out
        return detector_size / fringe_size
    

    def summary(self            : Propagator, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        plane_out = 'pupil' if self.inverse else 'focal'
            
        pixel_scale = convert_angular(self.pixel_scale_out, 'radians', 
                                        angular_units)
        shift = convert_angular(self.shift, 'radians', angular_units) \
            if not self.pixel_shift else self.shift

        prop_string = (f"Propagates the Wavefront to a {plane_out} plane "
                       f"with {self.npixels_out}x{self.npixels_out} pixels of "
                       f"size {pixel_scale:.{sigfigs}} {angular_units} ")
        
        if (self.shift == np.zeros(2)).all():
            return prop_string +  "using a 2-sided MFT."
        else:
            if self.pixel_shift:
                shift_string = (f"shifted by {shift} pixels in the (x, y) "
                                "dimension ")
            else:
                shift_string = (f"shifted by {shift} {angular_units} in the "
                                "(x, y) dimension ")
            return prop_string + shift_string + "using a 2-sided MFT."


class CartesianFFT(CartesianPropagator, FixedSamplingPropagator):
    """
    A Propagator class designed to propagate a wavefront to a plane using a
    Fast Fourier Transfrom, with the pixel scale units defined in meters/pixel.

    Attributes
    ----------
    focal_length : Array, meters
        The focal_length of the lens/mirror this propagator represents.
    inverse : bool
        Is this an 'inverse' propagation. Non-inverse propagations represents
        propagation from a pupil to a focal plane, and inverse represents
        propagation from a focal to a pupil plane.
    name : str
        The name for this propagator.
    """


    def __init__(self         : Propagator,
                 focal_length : Array,
                 inverse      : bool = False,
                 name         : str  = 'CartesianFFT') -> Propagator:
        """
        Parameters
        ----------
        focal_length : Array, meters
            The focal_length of the lens/mirror this propagator represents.
        inverse : bool = False
            Is this an 'inverse' propagation. Non-inverse propagations
            represents propagation from a pupil to a focal plane, and inverse
            represents propagation from a focal to a pupil plane.
        name : str = 'CartesianFFT'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name         = name,
                         inverse      = inverse,
                         focal_length = focal_length)


    def get_pixel_scale_out(self : Propagator, wavefront : Wavefront) -> Array:
        """
        The pixel scale in the output plane.

        Parameters
        ----------
        wavefront : Wavefront
            The `Wavefront` that is being propagted.

        Returns
        -------
        pixel_scale_out : Array, meters/pixel
            The pixel scale in the output plane.
        """
        return self.focal_length * wavefront.wavelength / wavefront.diameter
    

    def summary(self            : Propagator, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        plane_out = 'pupil' if self.inverse else 'focal'
        focal_length = convert_cartesian(self.focal_length, 'meters', 
                                        cartesian_units)

        return (f"Propagates the Wavefront {focal_length} {cartesian_units} to "
                f"a {plane_out} plane using a FFT.")


class AngularFFT(AngularPropagator, FixedSamplingPropagator):
    """
    A Propagator class designed to propagate a wavefront to a plane using a
    Fast Fourier Transfrom, with the pixel scale units defined in meters/pixel
    in pupil planes and radians/pixel in focal planes.

    Attributes
    ----------
    inverse : bool
        Is this an 'inverse' propagation. Non-inverse propagations represents
        propagation from a pupil to a focal plane, and inverse represents
        propagation from a focal to a pupil plane.
    name : str
        The name for this propagator.
    """


    def __init__(self    : Propagator,
                 inverse : bool = False,
                 name    : str = 'AngularFFT') -> Propagator:
        """
        Constructor for the AngularFFT propagator.

        Parameters
        ----------
        inverse : bool = False
            Is this an 'inverse' propagation. Non-inverse propagations
            represents propagation from a pupil to a focal plane, and inverse
            represents propagation from a focal to a pupil plane.
        name : str = 'AngularFFT'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name = name, inverse = inverse)


    def get_pixel_scale_out(self : Propagator, wavefront : Wavefront) -> Array:
        """
        Calculate the pixel scale in the output plane in units of radians per
        pixel.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront that is being propagated.

        Returns
        -------
        pixel_scale_out : Array, radians/pixel
            The pixel scale in the output plane.
        """
        return wavefront.wavelength / wavefront.diameter
    

    def summary(self            : Propagator, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        plane_out = 'pupil' if self.inverse else 'focal'

        return f"Propagates the Wavefront to a {plane_out} plane using a FFT."


class CartesianFresnel(FarFieldFresnel, CartesianMFT):
    """
    A propagator class to for Far-Field fresnel propagations. This classes
    implements algorithms that use quadratic phase factors to better represent
    out-of-plane behaviour of wavefronts, close to the focal plane. This class
    is designed to work on Cartesian wavefronts, ie pixel units are in
    meters/pixel in the output plane.

    Attributes
    ----------
    npixels_out : int
        The number of pixels in the output plane.
    pixel_scale_out : Array, meters/pixel
        The pixel scale in the output plane, measured in meters per pixel.
    focal_length : Array, meters
        The focal_length of the lens/mirror this propagator represents.
    propagation_shift : Array, meters
        The shift in the propagation distance of the wavefront.
    inverse : bool
        Is this an 'inverse' propagation. Non-inverse propagations represents
        propagation from a pupil to a focal plane, and inverse represents
        propagation from a focal to a pupil plane.
    shift : Array
        The (x, y) shift to apply to the wavefront in the output plane.
    pixel_shift : bool
        Should the shift value be considered in units of pixels, or in the
        physical units of the output plane (ie pixels or meters, radians). True
        interprets the shift value in pixel units.
    name : str
        The name for this propagator.
    """


    def __init__(self              : Propagator,
                 npixels_out       : Array,
                 pixel_scale_out   : Array,
                 focal_length      : Array,
                 propagation_shift : Array,
                 inverse           : bool  = False,
                 shift             : Array = np.array([0., 0.]),
                 pixel_shift       : bool  = False,
                 name              : str   = 'CartesianFresnel') -> Propagator:
        """
        Constructor for the CartesianFresnel propagator

        Parameters
        ----------
        pixel_scale_out : Array, meters/pixel
            The pixel scale in the output plane, measured in meters per pixel.
        npixels_out : int
            The number of pixels in the output plane.
        focal_length : Array, meters
            The focal_length of the lens/mirror this propagator represents.
        propagation_shift : Array, meters
            The shift in the propagation distance of the wavefront.
        inverse : bool = False
            Is this an 'inverse' propagation. Non-inverse propagations
            represents propagation from a pupil to a focal plane, and inverse
            represents propagation from a focal to a pupil plane.
        shift : Array = np.array([0., 0.])
            The (x, y) shift to apply to the wavefront in the output plane.
        pixel_shift : bool = False
            Should the shift value be considered in units of pixel, or in the
            physical units of the output plane (ie pixels or meters, radians).
        name : str = 'CartesianFresnel'
            The name of the layer, which is used to index the layers dictionary.
        """
        super().__init__(name               = name,
                         inverse            = inverse,
                         shift              = shift,
                         pixel_shift        = pixel_shift,
                         focal_length       = focal_length,
                         pixel_scale_out    = pixel_scale_out,
                         npixels_out        = npixels_out,
                         propagation_shift  = propagation_shift)


    def get_nfringes(self      : Propagator,
                     wavefront : Wavefront) -> Array:
        """
        The number of diffraction fringes in the output plane.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront being propagated.

        Returns
        -------
        fringes : Array
            The number of diffraction fringes in the output plane.
        """
        propagation_distance = self.focal_length + self.propagation_shift
        size_in = wavefront.diameter
        size_out = self.pixel_scale_out * self.npixels_out
        return size_in * size_out / wavefront.wavelength / propagation_distance


    def quadratic_phase(self          : Propagator,
                        x_coordinates : Array,
                        y_coordinates : Array,
                        wavelength    : Array,
                        distance      : Array) -> Array:
        """
        A convinience function for calculating quadratic phase factors.

        Parameters
        ----------
        x_coordinates : Array
            The x coordinates of the pixels in meters. This will be different
            in the plane of propagation and the initial plane.
        y_coordinates : Array
            The y coordinates of the pixels in meters. This will be different
            in the plane of propagation and the initial plane.
        wavelength : Array, meters
            The wavelength of the wavefront.
        distance : Array, meters
            The distance that is to be propagated in meters.

        Returns
        -------
        quadratic_phase : Array
            A set of phase factors that are useful in optical calculations.
        """
        wavenumber = 2 * np.pi / wavelength
        radial_coordinates = np.hypot(x_coordinates, y_coordinates)
        return np.exp(0.5j * wavenumber * radial_coordinates ** 2 / distance)


    def transfer_function(self      : Propagator,
                          wavefront : Wavefront,
                          distance  : Array) -> Array:
        """
        The Optical Transfer Function defining the phase evolution of the
        wavefront when propagating to a non-conjugate plane.

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.
        distance : Array, meters
            The distance that is being propagated in meters.

        Returns
        -------
        field : Array
            The field that represents the optical transfer.
        """
        return np.exp(1.0j * wavefront.wavenumber * distance)


    def propagate(self : Propagator, wavefront : Wavefront) -> Array:
        """
        Propagates the wavefront from the input plane to the output plane using
        a Matrix Fourier Transform.

        TODO: Set plane type to intermediate

        Parameters
        ----------
        wavefront : Wavefront
            The wavefront to propagate.

        Returns
        -------
        field : Array
            The normalised electric field phasor after the propagation.
        """
        # See gihub issue #52
        offsets = self.get_shift()

        input_positions = wavefront.pixel_coordinates
        output_positions = get_pixel_positions(
                            (self.npixels_out, self.npixels_out), 
                            (self.pixel_scale_out, self.pixel_scale_out))

        propagation_distance = self.focal_length + self.propagation_shift

        field = wavefront.phasor
        field *= self.quadratic_phase(*input_positions,
            wavefront.wavelength, - self.focal_length)
        field *= self.quadratic_phase(*input_positions,
            wavefront.wavelength, propagation_distance)
        wavefront = wavefront.set_phasor(np.abs(field), np.angle(field))

        field = super().propagate(wavefront)
        field *= self.transfer_function(wavefront, propagation_distance)
        field *= self.quadratic_phase(*output_positions,
            wavefront.wavelength, propagation_distance)
        return field
    

    def summary(self            : Propagator, 
                angular_units   : str = 'radians', 
                cartesian_units : str = 'meters', 
                sigfigs         : int = 4) -> str:
        """
        Returns a summary of the class.

        Parameters
        ----------
        angular_units : str = 'radians'
            The angular units to use in the summary. Options are 'radians', 
            'degrees', 'arcseconds' and 'arcminutes'.
        cartesian_units : str = 'meters'
            The cartesian units to use in the summary. Options are 'meters',
            'millimeters' and 'microns'.
        sigfigs : int = 4
            The number of significant figures to use in the summary.

        Returns
        -------
        summary : str
            A summary of the class.
        """
        plane_out = 'pupil' if self.inverse else 'focal'

        pixel_scale = convert_cartesian(self.pixel_scale_out, 'meters', 
                                        cartesian_units)
        prop_distance = convert_cartesian(self.focal_length, 'meters', 
                                        cartesian_units)
        defocus = convert_cartesian(self.propagation_shift, 'meters', 
                                        cartesian_units)
        shift = convert_cartesian(self.shift, 'meters', cartesian_units) \
            if not self.pixel_shift else self.shift


        prop_string = (f"Propagates the Wavefront {prop_distance} "
                       f"{cartesian_units} to a {plane_out} plane "
                       f"with {self.npixels_out}x{self.npixels_out} pixels of "
                       f"size {pixel_scale:.{sigfigs}} {cartesian_units} ")
        
        defocus_string = (f"with a defocus of {defocus} {cartesian_units} "
                           "using a 2-sided MFT.")
        
        if (self.shift == np.zeros(2)).all():
            return prop_string +  defocus_string
        else:
            if self.pixel_shift:
                shift_string = (f"shifted by {shift} pixels in the (x, y) "
                                "dimension ")
            else:
                shift_string = (f"shifted by {shift} {cartesian_units} in the "
                                "(x, y) dimension ")
            return prop_string + shift_string + defocus_string


# TODO: Implement eventually
# class AngularFresnel(FarFieldFresnel, AngularMFT):
#     """
#     Propagates an AngularWavefront in the Fresnel approximation.

#     Attributes
#     ----------
#     """
#     pass